#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.linalg import LinAlgError

from .... import operands
from ...core import ExecutableTuple
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor


class TensorLU(operands.LU, TensorOperandMixin):
    def __init__(self, dtype=None, **kw):
        super(TensorLU, self).__init__(_dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super(TensorLU, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        import scipy.linalg

        a = astensor(a)
        if a.ndim != 2:
            raise LinAlgError('{0}-dimensional array given. '
                              'Tensor must be two-dimensional'.format(a.ndim))
        if a.shape[0] != a.shape[1]:
            raise LinAlgError('Input must be square')

        tiny_p, tiny_l, tiny_u = scipy.linalg.lu(np.array([[1, 2], [2, 5]], dtype=a.dtype))

        p, l, u = self.new_tensors([a], (a.shape, a.shape, a.shape),
                                   kws=[
                                       {'side': 'p', 'dtype': tiny_p.dtype},
                                       {'side': 'l', 'dtype': tiny_l.dtype},
                                       {'side': 'u', 'dtype': tiny_u.dtype},
                                   ])
        return ExecutableTuple([p, l, u])

    @classmethod
    def tile(cls, op):
        from ..arithmetic.subtract import TensorSubtract
        from ..arithmetic.utils import tree_add
        from ..base.transpose import TensorTranspose
        from ..datasource.zeros import TensorZeros
        from .dot import TensorDot
        from .solve_triangular import TensorSolveTriangular

        P, L, U = op.outputs
        in_tensor = op.input
        if len(set(in_tensor.nsplits[0] + in_tensor.nsplits[1])) != 1:
            nsplit = min(in_tensor.nsplits[0] + in_tensor.nsplits[1])
            # input's chunks must be all square
            in_tensor = in_tensor.rechunk([nsplit for _ in range(in_tensor.ndim)]).single_tiles()
            if len(set(in_tensor.nsplits[0] + in_tensor.nsplits[1])) != 1:
                raise LinAlgError('All chunks must be a square matrix to perform LU decomposition.')

        p_chunks, p_invert_chunks, lower_chunks, l_permuted_chunks, upper_chunks = {}, {}, {}, {}, {}
        for i in range(in_tensor.chunk_shape[0]):
            for j in range(in_tensor.chunk_shape[1]):
                if i < j:
                    chunk_shape = (in_tensor.nsplits[0][i], in_tensor.nsplits[1][j])
                    p_chunk = TensorZeros().new_chunk(None, chunk_shape, index=(i, j))
                    lower_chunk = TensorZeros().new_chunk(None, chunk_shape, index=(i, j))
                    p_chunks[p_chunk.index] = p_chunk
                    lower_chunks[lower_chunk.index] = lower_chunk

                    target_u = in_tensor.cix[i, j]
                    p_invert = p_invert_chunks[i, i]
                    target = TensorDot(dtype=U.op.dtype).new_chunk(
                        [p_invert, target_u], (p_invert.shape[0], target_u.shape[1]))
                    if i > 0:
                        prev_chunks_u = []
                        for p in range(i):
                            a, b = lower_chunks[i, p], upper_chunks[p, j]
                            prev_chunk = TensorDot(dtype=U.op.dtype).new_chunk(
                                [a, b], (a.shape[0], b.shape[1]))
                            prev_chunks_u.append(prev_chunk)
                        if len(prev_chunks_u) == 1:
                            s = prev_chunks_u[0]
                        else:
                            s = tree_add(prev_chunks_u[0].dtype, prev_chunks_u,
                                         None, prev_chunks_u[0].shape)
                        target = TensorSubtract(dtype=U.op.dtype).new_chunk(
                            [target, s, None, None], target.shape)
                    upper_chunk = TensorSolveTriangular(lower=True, dtype=U.op.dtype).new_chunk(
                        [lower_chunks[i, i], target], target.shape, index=(i, j))
                    upper_chunks[upper_chunk.index] = upper_chunk
                elif i == j:
                    target = in_tensor.cix[i, j]
                    if i > 0:
                        prev_chunks = []
                        for p in range(i):
                            a, b = l_permuted_chunks[i, p], upper_chunks[p, j]
                            prev_chunk = TensorDot(dtype=op.dtype).new_chunk(
                                [a, b], (a.shape[0], b.shape[1]))
                            prev_chunks.append(prev_chunk)
                        if len(prev_chunks) == 1:
                            s = prev_chunks[0]
                        else:
                            s = tree_add(prev_chunks[0].dtype, prev_chunks,
                                         None, prev_chunks[0].shape)
                        target = TensorSubtract(dtype=op.dtype).new_chunk([target, s, None, None], target.shape)
                    new_op = TensorLU(dtype=op.dtype)
                    lu_chunks = new_op.new_chunks([target], (target.shape, target.shape, target.shape),
                                                  index=(i, j),
                                                  kws=[{'side': 'p'}, {'side': 'l'}, {'side': 'u'}])
                    p_chunk, lower_chunk, upper_chunk = lu_chunks
                    # transposed p equals to inverted p
                    p_chunk_invert = TensorTranspose(dtype=p_chunk.dtype).new_chunk(
                        [p_chunk], p_chunk.shape, index=p_chunk.index)
                    p_chunks[p_chunk.index] = p_chunk
                    p_invert_chunks[p_chunk_invert.index] = p_chunk_invert
                    lower_chunks[lower_chunk.index] = lower_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk

                    # l_permuted should be transferred to the final lower triangular
                    for p in range(i):
                        l_permuted_chunk = l_permuted_chunks[i, p]
                        l_chunk = TensorDot(dtype=L.op.dtype).new_chunk(
                            [p_chunk_invert, l_permuted_chunk],
                            (p_chunk_invert.shape[0], l_permuted_chunk.shape[1]),
                            index=l_permuted_chunk.index
                        )
                        lower_chunks[l_permuted_chunk.index] = l_chunk
                else:
                    chunk_shape = (in_tensor.nsplits[0][i], in_tensor.nsplits[1][j])
                    p_chunk = TensorZeros().new_chunk(None, chunk_shape, index=(i, j))
                    upper_chunk = TensorZeros().new_chunk(None, chunk_shape, index=(i, j))
                    p_chunks[p_chunk.index] = p_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk
                    target_l = in_tensor.cix[i, j]
                    if j > 0:
                        prev_chunks_l = []
                        for p in range(j):
                            a, b = l_permuted_chunks[i, p], upper_chunks[p, j]
                            prev_chunk = TensorDot(dtype=L.op.dtype).new_chunk(
                                [a, b], (a.shape[0], b.shape[1]))
                            prev_chunks_l.append(prev_chunk)
                        if len(prev_chunks_l) == 1:
                            s = prev_chunks_l[0]
                        else:
                            s = tree_add(prev_chunks_l[0].dtype, prev_chunks_l,
                                         None, prev_chunks_l[0].shape)
                        target_l = TensorSubtract(dtype=L.op.dtype).new_chunk(
                            [target_l, s, None, None], target_l.shape)
                    u = upper_chunks[j, j]
                    a_transpose = TensorTranspose(dtype=u.dtype).new_chunk([u], u.shape)
                    target_transpose = TensorTranspose(dtype=target_l.dtype).new_chunk([target_l], target_l.shape)
                    lower_permuted_chunk = TensorSolveTriangular(lower=True, dtype=L.op.dtype).new_chunk(
                        [a_transpose, target_transpose], target_l.shape, index=(i, j))
                    lower_transpose = TensorTranspose(dtype=lower_permuted_chunk.dtype).new_chunk(
                        [lower_permuted_chunk], lower_permuted_chunk.shape, index=lower_permuted_chunk.index)
                    l_permuted_chunks[lower_permuted_chunk.index] = lower_transpose

        new_op = op.copy()
        kws = [
            {'chunks': list(p_chunks.values()), 'nsplits': in_tensor.nsplits, 'dtype': P.dtype},
            {'chunks': list(lower_chunks.values()), 'nsplits': in_tensor.nsplits, 'dtype': L.dtype},
            {'chunks': list(upper_chunks.values()), 'nsplits': in_tensor.nsplits, 'dtype': U.dtype}
        ]
        return new_op.new_tensors(op.inputs, [P.shape, L.shape, U.shape], kws=kws)


def lu(a):
    """
    LU decomposition

    The decomposition is::
        A = P L U
    where P is a permutation matrix, L lower triangular with unit diagonal elements,
    and U upper triangular.

    Parameters
    ----------
    a : (M, N) array_like
        Array to decompose

    Returns
    -------
    p : (M, M) ndarray
        Permutation matrix
    l : (M, K) ndarray
        Lower triangular or trapezoidal matrix with unit diagonal.
        K = min(M, N)
    u : (K, N) ndarray
        Upper triangular or trapezoidal matrix

    Examples
    --------
    >>> import mars.tensor as mt

    >>> A = mt.array([[1,2],[2,3]])
    >>> A.execute()
    array([[ 1,  2],
           [ 2,  3]])
    >>> P, L, U = mt.linalg.lu(A)
    >>> P.execute()
    array([[ 0,  1],
           [ 1,  0]])
    >>> L.execute()
    array([[ 1,  0],
           [ 0.5,  1]])
    >>> U.execute()
    array([[ 2,  3],
           [ 0,  0.5]])
    >>> mt.dot(P.dot(L), U).execute() # verify that PL * U = A
    array([[ 1,  2],
           [ 2,  3]])

    """
    op = TensorLU()
    return op(a)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/tensor.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mars.serialize.protos import value_pb2 as mars_dot_serialize_dot_protos_dot_value__pb2
from mars.serialize.protos import chunk_pb2 as mars_dot_serialize_dot_protos_dot_chunk__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/tensor.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\"mars/serialize/protos/tensor.proto\x1a!mars/serialize/protos/value.proto\x1a!mars/serialize/protos/chunk.proto\"\xaa\x01\n\tTensorDef\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05shape\x18\x02 \x03(\x03\x12\x15\n\x05\x64type\x18\x03 \x01(\x0b\x32\x06.Value\x12\x12\n\x02op\x18\x04 \x01(\x0b\x32\x06.Value\x12\x17\n\x07nsplits\x18\x05 \x01(\x0b\x32\x06.Value\x12\x19\n\x06\x63hunks\x18\x06 \x03(\x0b\x32\t.ChunkDef\x12\x16\n\x06params\x18\x07 \x01(\x0b\x32\x06.Value\x12\n\n\x02id\x18\x08 \x01(\tb\x06proto3')
  ,
  dependencies=[mars_dot_serialize_dot_protos_dot_value__pb2.DESCRIPTOR,mars_dot_serialize_dot_protos_dot_chunk__pb2.DESCRIPTOR,])




_TENSORDEF = _descriptor.Descriptor(
  name='TensorDef',
  full_name='TensorDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='TensorDef.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shape', full_name='TensorDef.shape', index=1,
      number=2, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dtype', full_name='TensorDef.dtype', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='op', full_name='TensorDef.op', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nsplits', full_name='TensorDef.nsplits', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chunks', full_name='TensorDef.chunks', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='params', full_name='TensorDef.params', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='TensorDef.id', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=109,
  serialized_end=279,
)

_TENSORDEF.fields_by_name['dtype'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_TENSORDEF.fields_by_name['op'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_TENSORDEF.fields_by_name['nsplits'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_TENSORDEF.fields_by_name['chunks'].message_type = mars_dot_serialize_dot_protos_dot_chunk__pb2._CHUNKDEF
_TENSORDEF.fields_by_name['params'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
DESCRIPTOR.message_types_by_name['TensorDef'] = _TENSORDEF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TensorDef = _reflection.GeneratedProtocolMessageType('TensorDef', (_message.Message,), dict(
  DESCRIPTOR = _TENSORDEF,
  __module__ = 'mars.serialize.protos.tensor_pb2'
  # @@protoc_insertion_point(class_scope:TensorDef)
  ))
_sym_db.RegisterMessage(TensorDef)


# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ...core import build_mode
from ..utils import inject_dtype
from ..datasource import tensor as astensor
from .core import TensorCompare, TensorCompareConstant


class TensorEqual(operands.Equal, TensorCompare):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorEqual, self).__init__(_casting=casting, _err=err,
                                          _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def constant_cls(cls):
        return TensorEqConstant


class TensorEqConstant(operands.EqConstant, TensorCompareConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorEqConstant, self).__init__(_casting=casting, _err=err,
                                               _dtype=dtype, _sparse=sparse, **kw)


@inject_dtype(np.bool_)
def equal(x1, x2, out=None, where=None, **kwargs):
    """
    Return (x1 == x2) element-wise.

    Parameters
    ----------
    x1, x2 : array_like
        Input tensors of the same shape.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated array is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs
        For other keyword-only arguments, see the
        :ref:`ufunc docs <ufuncs.kwargs>`.

    Returns
    -------
    out : Tensor or bool
        Output tensor of bools, or a single bool if x1 and x2 are scalars.

    See Also
    --------
    not_equal, greater_equal, less_equal, greater, less

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.equal([0, 1, 3], mt.arange(3)).execute()
    array([ True,  True, False])

    What is compared are values, not types. So an int (1) and a tensor of
    length one can evaluate as True:

    >>> mt.equal(1, mt.ones(1))
    array([ True])
    """
    if build_mode().is_build_mode:
        return astensor(x1)._equals(x2)

    op = TensorEqual(**kwargs)
    return op(x1, x2, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def result_type(*tensors_and_dtypes):
    """
    Returns the type that results from applying the NumPy
    type promotion rules to the arguments.

    Type promotion in Mars works similarly to the rules in languages
    like C++, with some slight differences.  When both scalars and
    arrays are used, the array's type takes precedence and the actual value
    of the scalar is taken into account.

    For example, calculating 3*a, where a is an array of 32-bit floats,
    intuitively should result in a 32-bit float output.  If the 3 is a
    32-bit integer, the NumPy rules indicate it can't convert losslessly
    into a 32-bit float, so a 64-bit float should be the result type.
    By examining the value of the constant, '3', we see that it fits in
    an 8-bit integer, which can be cast losslessly into the 32-bit float.

    Parameters
    ----------
    tensors_and_dtypes : list of tensors and dtypes
        The operands of some operation whose result type is needed.

    Returns
    -------
    out : dtype
        The result type.

    See also
    --------
    dtype, promote_types, min_scalar_type, can_cast

    Notes
    -----
    The specific algorithm used is as follows.

    Categories are determined by first checking which of boolean,
    integer (int/uint), or floating point (float/complex) the maximum
    kind of all the arrays and the scalars are.

    If there are only scalars or the maximum category of the scalars
    is higher than the maximum category of the arrays,
    the data types are combined with :func:`promote_types`
    to produce the return value.

    Otherwise, `min_scalar_type` is called on each array, and
    the resulting data types are all combined with :func:`promote_types`
    to produce the return value.

    The set of int values is not a subset of the uint values for types
    with the same number of bits, something not reflected in
    :func:`min_scalar_type`, but handled as a special case in `result_type`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.result_type(3, mt.arange(7, dtype='i1'))
    dtype('int8')

    >>> mt.result_type('i4', 'c8')
    dtype('complex128')

    >>> mt.result_type(3.0, -2)
    dtype('float64')
    """
    from ...core import Tensor

    arrays_and_dtypes = [a.dtype if isinstance(a, Tensor) else a
                         for a in tensors_and_dtypes]
    return np.result_type(*arrays_and_dtypes)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import json
import time

from ...api import MarsAPI
from ...graph import DirectedGraph
from ...scheduler.graph import GraphState
from ...serialize import dataserializer


class LocalClusterSession(object):
    def __init__(self, endpoint):
        self._session_id = uuid.uuid4()
        self._endpoint = endpoint
        self._tensor_to_graph = dict()
        self._api = MarsAPI(self._endpoint)

        # create session on the cluster side
        self._api.create_session(self._session_id)

    def run(self, *tensors, **kw):
        timeout = kw.pop('timeout', -1)
        if kw:
            raise TypeError('run got unexpected key arguments {0}'.format(', '.join(kw.keys())))

        graph = DirectedGraph()
        targets = [t.key for t in tensors]
        graph_key = uuid.uuid4()
        for t in tensors:
            t.build_graph(graph, tiled=False)
            self._tensor_to_graph[t.key] = graph_key

        # submit graph to local cluster
        self._api.submit_graph(self._session_id, json.dumps(graph.to_json()),
                               graph_key, targets)

        exec_start_time = time.time()
        while timeout <= 0 or time.time() - exec_start_time <= timeout:
            time.sleep(0.1)

            graph_state = self._api.get_graph_state(self._session_id, graph_key)
            if graph_state == GraphState.SUCCEEDED:
                break
            if graph_state == GraphState.FAILED:
                # TODO(qin): add traceback
                raise SystemError('Graph execution failed with unknown reason')

        if 0 < timeout < time.time() - exec_start_time:
            raise TimeoutError

        data_list = []
        for target in targets:
            resp = self._api.fetch_data(self._session_id, graph_key, target)
            data_list.append(dataserializer.loads(resp))

        return data_list

    def decref(self, *keys):
        for k in keys:
            if k not in self._tensor_to_graph:
                continue
            graph_key = self._tensor_to_graph[k]
            self._api.delete_data(self._session_id, graph_key, k)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self._api.delete_session(self._session_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.linalg import LinAlgError

from .... import operands
from ...core import ExecutableTuple
from ..datasource import tensor as astensor
from .core import TSQR


class TensorQR(operands.QR, TSQR):
    def __init__(self, method=None, dtype=None, **kw):
        super(TensorQR, self).__init__(_method=method, _dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super(TensorQR, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        a = astensor(a)

        if a.ndim != 2:
            raise LinAlgError('{0}-dimensional tensor given. '
                              'Tensor must be two-dimensional'.format(a.ndim))

        tiny_q, tiny_r = np.linalg.qr(np.ones((1, 1), dtype=a.dtype))

        x, y = a.shape
        q_shape = a.shape if x > y else (x, x)
        r_shape = a.shape if x < y else (y, y)
        q, r = self.new_tensors([a], (q_shape, r_shape),
                                kws=[{'side': 'q', 'dtype': tiny_q.dtype},
                                     {'side': 'r', 'dtype': tiny_r.dtype}])
        return ExecutableTuple([q, r])

    @classmethod
    def tile(cls, op):
        q, r = op.outputs
        q_dtype, r_dtype = cls._get_obj_attr(q, 'dtype'), cls._get_obj_attr(r, 'dtype')
        q_shape, r_shape = cls._get_obj_attr(q, 'shape'), cls._get_obj_attr(r, 'shape')
        in_tensor = op.input
        if in_tensor.chunk_shape == (1, 1):
            in_chunk = in_tensor.chunks[0]
            chunk_op = op.copy().reset_key()
            qr_chunks = chunk_op.new_chunks([in_chunk], (q_shape, r_shape), index=in_chunk.index,
                                            kws=[{'side': 'q'}, {'side': 'r'}])
            q_chunk, r_chunk = qr_chunks

            new_op = op.copy()
            kws = [
                {'chunks': [q_chunk], 'nsplits': ((1,), (1,)), 'dtype': q_dtype},
                {'chunks': [r_chunk], 'nsplits': ((1,), (1,)), 'dtype': r_dtype}
            ]
            return new_op.new_tensors(op.inputs, [q_shape, r_shape], kws=kws)
        elif op.method == 'tsqr':
            return super(TensorQR, cls).tile(op)
        # TODO(hks): support sfqr(short-and-fat qr)
        else:
            raise NotImplementedError('Only tsqr method supported for now')


def qr(a, method='tsqr'):
    """
    Compute the qr factorization of a matrix.

    Factor the matrix `a` as *qr*, where `q` is orthonormal and `r` is
    upper-triangular.

    Parameters
    ----------
    a : array_like, shape (M, N)
        Matrix to be factored.
    method: {'tsqr'}, optional
        method to calculate qr factorization, tsqr as default

        TSQR is presented in:

            A. Benson, D. Gleich, and J. Demmel.
            Direct QR factorizations for tall-and-skinny matrices in
            MapReduce architectures.
            IEEE International Conference on Big Data, 2013.
            http://arxiv.org/abs/1301.1071

    Returns
    -------
    q : Tensor of float or complex, optional
        A matrix with orthonormal columns. When mode = 'complete' the
        result is an orthogonal/unitary matrix depending on whether or not
        a is real/complex. The determinant may be either +/- 1 in that
        case.
    r : Tensor of float or complex, optional
        The upper-triangular matrix.

    Raises
    ------
    LinAlgError
        If factoring fails.

    Notes
    -----
    For more information on the qr factorization, see for example:
    http://en.wikipedia.org/wiki/QR_factorization

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.random.randn(9, 6)
    >>> q, r = mt.linalg.qr(a)
    >>> mt.allclose(a, mt.dot(q, r)).execute()  # a does equal qr
    True

    """
    op = TensorQR(method=method)
    return op(a)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
import base64
import copy
import logging
import time

from ..config import options
from ..compat import six
from ..errors import *
from ..utils import log_unhandled
from .assigner import AssignerActor
from .kvstore import KVStoreActor
from .graph import GraphActor
from .resource import ResourceActor
from .utils import SchedulerActor, OperandState, GraphState, array_to_bytes

logger = logging.getLogger(__name__)


class OperandActor(SchedulerActor):
    """
    Actor handling the whole lifecycle of a particular operand instance
    """
    @staticmethod
    def gen_uid(session_id, op_key):
        return 's:operator$%s$%s' % (session_id, op_key)

    def __init__(self, session_id, graph_id, op_key, op_info, worker_endpoint=None,
                 is_terminal=False):
        super(OperandActor, self).__init__()
        op_info = copy.deepcopy(op_info)

        self._session_id = session_id
        self._graph_id = graph_id
        self._op_key = op_key
        self._op_path = '/sessions/%s/operands/%s' % (self._session_id, self._op_key)

        self._cluster_info_ref = None
        self._assigner_ref = None
        self._graph_ref = None
        self._resource_ref = None
        self._kv_store_ref = None

        self._op_name = op_info['op_name']
        self._state = OperandState(op_info['state'].lower())
        self._last_state = self._state
        self._retries = op_info['retries']
        self._is_terminal = is_terminal
        self._io_meta = op_info['io_meta']
        self._pred_keys = self._io_meta['predecessors']
        self._succ_keys = self._io_meta['successors']
        self._input_chunks = self._io_meta['input_chunks']
        self._chunks = self._io_meta['chunks']
        self._info = op_info

        # worker the operand expected to be executed on
        self._expect_worker = op_info.get('target_worker')
        # worker actually assigned
        self._worker_endpoint = worker_endpoint

        # ref of ExecutionActor on worker
        self._execution_ref = None
        # set of finished predecessors, used to decide whether we should move the operand to ready
        self._finish_preds = set()
        # set of finished successors, used to detect whether we can do clean up
        self._finish_succs = set()

        self._last_demand_depths = self._info.get('optimize', {}).get('demand_depths', ())

        self._input_worker_scores = dict()
        self._worker_scores = dict()

        # handlers of states. will be called when the state of the operand switches
        # from one to another
        self._state_handlers = {
            OperandState.READY: self._handle_ready,
            OperandState.RUNNING: self._handle_running,
            OperandState.FINISHED: self._handle_finished,
            OperandState.FREED: self._handle_freed,
            OperandState.FATAL: self._handle_fatal,
            OperandState.CANCELLING: self._handle_cancelling,
            OperandState.CANCELLED: self._handle_cancelled,
        }

    def post_create(self):
        self.set_cluster_info_ref()
        self._assigner_ref = self.get_promise_ref(AssignerActor.gen_name(self._session_id))
        self._graph_ref = self.get_actor_ref(GraphActor.gen_name(self._session_id, self._graph_id))
        self._resource_ref = self.get_actor_ref(ResourceActor.default_name())
        self._kv_store_ref = self.get_actor_ref(KVStoreActor.default_name())

    def add_finished_predecessor(self, op_key):
        self._finish_preds.add(op_key)
        if all(k in self._finish_preds for k in self._pred_keys):
            # all predecessors done, the operand can be executed now
            self.state = OperandState.READY
            self.start_operand()
            return True
        self.update_demand_depths(self._info.get('optimize', {}).get('depth', 0))
        return False

    def add_finished_successor(self, op_key):
        self._finish_succs.add(op_key)
        if not self._is_terminal and all(k in self._finish_succs for k in self._succ_keys):
            # non-terminal operand with all successors done, the data can be freed
            self.ref().free_data(_tell=True)

    def update_demand_depths(self, depth):
        """
        Update the depth of operand demanding data, or demanding the descendant of data
        produced by the current operand
        :param depth: depth to update
        """
        demand_depths = list(self._info.get('optimize', {}).get('demand_depths', ()))
        if not demand_depths:
            demand_depths = [depth]
        else:
            idx = 0
            for idx, v in enumerate(demand_depths):
                if v <= depth:
                    break
            if demand_depths[idx] == depth:
                return
            elif demand_depths[idx] > depth:
                demand_depths.append(depth)
            else:
                demand_depths.insert(idx, depth)
        if 'optimize' not in self._info:
            self._info['optimize'] = dict()
        self._info['optimize']['demand_depths'] = tuple(demand_depths)
        self._kv_store_ref.write('%s/optimize/demand_depths' % self._op_path,
                                 base64.b64encode(array_to_bytes('I', demand_depths)), _tell=True, _wait=False)
        if self.state == OperandState.READY:
            # if the operand is already submitted to AssignerActor, we need to update the priority
            self._assigner_ref.update_priority(self._op_key, self._info['optimize'], _tell=True, _wait=False)
        else:
            # send update command to predecessors
            for in_key in self._pred_keys:
                self._get_operator_actor(in_key).update_demand_depths(depth, _tell=True, _wait=False)

    def propose_descendant_workers(self, input_key, worker_scores, depth=1):
        """
        Calculate likelihood of the operand being sent to workers
        :param input_key: input key that carries the scores
        :param worker_scores: score of every worker on input key
        :param depth: maximal propagate depth
        """
        if self._worker_endpoint:
            # worker already assigned, there should be no other possibilities
            self._worker_scores = {self._worker_endpoint: 1.0}
        elif self._expect_worker:
            # worker already propsed, there should be no other possibilities
            self._worker_scores = {self._expect_worker: 1.0}
        else:
            # aggregate the score from input to the score of current operand
            old_scores = self._input_worker_scores.get(input_key, {})
            self._input_worker_scores[input_key] = worker_scores
            all_keys = set(old_scores.keys()) | set(worker_scores)
            for k in all_keys:
                delta = (worker_scores.get(k, 0) - old_scores.get(k, 0)) * 1.0 / len(self._pred_keys)
                self._worker_scores[k] = self._worker_scores.get(k, 0) + delta
                if self._worker_scores[k] < 1e-6:
                    del self._worker_scores[k]

        if depth:
            # push down to successors
            for succ_key in self._succ_keys:
                self._get_operator_actor(succ_key).propose_descendant_workers(
                    self._op_key, self._worker_scores, depth=depth - 1, _tell=True)
        # pick the worker with largest likelihood
        max_score = 0
        max_worker = None
        for k, v in self._worker_scores.items():
            if v > max_score:
                max_score = v
                max_worker = k
        if max_score > 0.5:
            logger.debug('Operand %s(%s) now owning a dominant worker %s. scores=%r',
                         self._op_key, self._op_name, max_worker, self._worker_scores)
            return self._input_chunks, max_worker

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._last_state = self._state
        if value != self._last_state:
            logger.debug('Operand %s(%s) state from %s to %s.', self._op_key, self._op_name,
                         self._last_state, value)
        self._state = value
        self._info['state'] = value.name
        self._graph_ref.set_operand_state(self._op_key, value.value, _tell=True)
        self._kv_store_ref.write('%s/state' % self._op_path, value.name, _tell=True)

    @property
    def retries(self):
        return self._retries

    @retries.setter
    def retries(self, value):
        self._retries = value
        self._info['retries'] = value
        self._kv_store_ref.write('%s/retries' % self._op_path, str(value), _tell=True)

        retry_timestamp = time.time()
        self._info['retry_timestamp'] = retry_timestamp
        self._kv_store_ref.write('%s/retry_timestamp' % self._op_path, str(value), _tell=True)

    def get_op_info(self):
        info = dict()
        info['name'] = self._op_name
        info['state'] = self.state
        return info

    def start_operand(self):
        """
        Start handling operand given self.state
        """
        self._state_handlers[self.state]()

    def stop_operand(self):
        """
        Stop operand by starting CANCELLING procedure
        """
        if self.state == OperandState.CANCELLING or self.state == OperandState.CANCELLED:
            return
        self.state = OperandState.CANCELLING
        self.start_operand()

    def _free_worker_data(self, ep, chunk_key):
        """
        Free data on single worker
        :param ep: worker endpoint
        :param chunk_key: chunk key
        """
        from ..worker.chunkholder import ChunkHolderActor

        worker_cache_ref = self.ctx.actor_ref(ep, ChunkHolderActor.default_name())
        worker_cache_ref.unregister_chunk(self._session_id, chunk_key, _tell=True)

    def free_data(self, state=OperandState.FREED):
        """
        Free output data of current operand
        :param state: target state
        """
        if self.state == OperandState.FREED:
            return
        for chunk_key in self._chunks:
            chunk_path = '/sessions/%s/chunks/%s' % (self._session_id, chunk_key)
            endpoints = self._kv_store_ref.read(chunk_path + '/workers', silent=True)
            if endpoints is None:
                continue
            for ep_item in endpoints.children:
                ep = ep_item.key.rsplit('/', 1)[-1]
                self._free_worker_data(ep, chunk_key)
            self._kv_store_ref.delete(chunk_path, dir=True, recursive=True, _tell=True)
        self.state = state
        self.start_operand()

    def propagate_state(self, state):
        """
        Change state of current operand by other operands
        :param state: target state
        """
        if self.state == OperandState.CANCELLING or self.state == OperandState.CANCELLED:
            return
        if self.state != state:
            self.state = state
            self.start_operand()

    def _get_raw_execution_ref(self):
        """
        Get raw ref of ExecutionActor on assigned worker. This method can be patched on debug
        """
        from ..worker.dispatcher import DispatchActor

        dispatch_ref = self.promise_ref(DispatchActor.default_name(), address=self._worker_endpoint)
        exec_uid = dispatch_ref.get_hash_slot('execution', self._op_key)
        return self.ctx.actor_ref(exec_uid, address=self._worker_endpoint)

    def _get_execution_ref(self):
        """
        Get ref of ExecutionActor on assigned worker
        """
        if self._execution_ref is None:
            self._execution_ref = self.promise_ref(self._get_raw_execution_ref())
        return self._execution_ref

    def _get_operator_actor(self, key):
        """
        Get ref of OperandActor by operand key
        """
        op_uid = self.gen_uid(self._session_id, key)
        return self.ctx.actor_ref(self.get_scheduler(op_uid), op_uid)

    def _get_multiple_chunk_size(self, chunk_keys):
        if not chunk_keys:
            return tuple()
        keys = ['/sessions/%s/chunks/%s/data_size' % (self._session_id, chunk_key)
                for chunk_key in chunk_keys]
        return (res.value for res in self._kv_store_ref.read_batch(keys))

    @log_unhandled
    def _handle_ready(self):
        @log_unhandled
        def _submit_job(worker):
            # worker assigned, submit job
            if self.state in (OperandState.CANCELLED, OperandState.CANCELLING):
                self.start_operand()
                return

            self._worker_endpoint = worker

            target_predicts = dict()
            if options.scheduler.enable_active_push:
                # if active push enabled, we calculate the most possible target
                for succ_key in self._succ_keys:
                    succ_worker_predict = self._get_operator_actor(succ_key) \
                        .propose_descendant_workers(self._op_key, {worker: 1.0})
                    if not succ_worker_predict:
                        continue
                    keys, target = succ_worker_predict
                    if target == worker:
                        continue
                    for k in keys:
                        if k not in self._chunks:
                            continue
                        if k not in target_predicts:
                            target_predicts[k] = set()
                        target_predicts[k].add(target)
            if not target_predicts:
                target_predicts = None
            else:
                logger.debug('Receive active pushing list for operand %s: %r',
                             self._op_key, target_predicts)

            data_sizes = dict(zip(
                self._input_chunks,
                [int(v) for v in self._get_multiple_chunk_size(self._input_chunks)],
            ))

            # submit job
            self._execution_ref = self._get_execution_ref()
            serialized_exec_graph = self._graph_ref.get_executable_operand_dag(self._op_key)
            self._execution_ref.execute_graph(self._session_id, self._op_key, serialized_exec_graph,
                                              self._io_meta, data_sizes, send_targets=target_predicts,
                                              _promise=True)
            self.state = OperandState.RUNNING
            self.start_operand()

        logger.debug('Applying for resources for operand %r', self._info)
        # if under retry, give application a delay
        delay = options.scheduler.retry_delay if self.retries else 0
        # Send resource application. Submit job when worker assigned
        self._assigner_ref.apply_for_resource(
            self._session_id, self._op_key, self._info, _delay=delay, _promise=True) \
            .then(_submit_job)

    @log_unhandled
    def _handle_running(self):
        self._execution_ref = self._get_execution_ref()

        @log_unhandled
        def _acceptor(*_):
            # handling success of operand execution
            self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
            self.state = OperandState.FINISHED
            self.start_operand()

        @log_unhandled
        def _rejecter(*exc):
            # handling execption occurance of operand execution
            self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
            if exc and not isinstance(exc[0], type):
                raise TypeError('Unidentified rejection args: %r', exc)
            if exc and issubclass(exc[0], ResourceInsufficient):
                # resource insufficient: just set to READY and continue
                self.state = OperandState.READY
                self.ref().start_operand(_tell=True)
            elif exc and issubclass(exc[0], ExecutionInterrupted):
                # job cancelled: switch to cancelled
                self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
                logger.warning('Execution of operand %s interrupted.', self._op_key)
                self.free_data(OperandState.CANCELLED)
            else:
                self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
                try:
                    if exc:
                        six.reraise(*exc)
                    else:
                        raise SystemError('Worker throws rejection without details')
                except:
                    logger.exception('Attempt %d: Unexpected error occurred in executing operand %s in %s',
                                     self.retries + 1, self._op_key, self._worker_endpoint)
                # increase retry times
                self.retries += 1
                if self.retries >= options.scheduler.retry_num:
                    # no further trial
                    self.state = OperandState.FATAL
                else:
                    self.state = OperandState.READY
                self.start_operand()

        self._execution_ref.add_finish_callback(self._session_id, self._op_key, _promise=True) \
            .then(_acceptor, _rejecter)

    @log_unhandled
    def _handle_finished(self):
        self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
        if self._last_state == OperandState.CANCELLING:
            self.state = OperandState.CANCELLING
            self.start_operand()
            return
        new_ready_generated = False
        # update pred & succ finish records to trigger further actions
        # record if successors can be executed
        for out_key in self._succ_keys:
            new_ready_generated |= self._get_operator_actor(out_key).add_finished_predecessor(self._op_key)
        for in_key in self._pred_keys:
            self._get_operator_actor(in_key).add_finished_successor(self._op_key)
        # require more chunks to execute if the completion caused no successors to run
        if not new_ready_generated:
            self._assigner_ref.allocate_top_resources(_tell=True)
        if self._is_terminal:
            # update records in GraphActor to help decide if the whole graph finished execution
            self._graph_ref.mark_terminal_finished(self._op_key, _tell=True)

    @log_unhandled
    def _handle_fatal(self):
        if self._last_state == OperandState.FATAL:
            return
        if self._worker_endpoint:
            self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)

        if self._is_terminal:
            # update records in GraphActor to help decide if the whole graph finished execution
            self._graph_ref.mark_terminal_finished(self._op_key, final_state=GraphState.FAILED, _tell=True)
        # set successors to FATAL
        for k in self._succ_keys:
            self._get_operator_actor(k).propagate_state(OperandState.FATAL, _tell=True)

    @log_unhandled
    def _handle_cancelling(self):
        if self._last_state == OperandState.CANCELLING:
            return
        if self._last_state == OperandState.CANCELLED:
            self.state = OperandState.CANCELLED
            return

        if self._last_state == OperandState.RUNNING:
            # send stop to worker
            self._execution_ref = self._get_execution_ref()
            self._execution_ref.stop_execution(self._op_key, _tell=True)
            return
        if self._last_state == OperandState.FINISHED:
            # delete data on cancelled
            self.ref().free_data(state=OperandState.CANCELLED, _tell=True)
            return
        if self._last_state == OperandState.READY:
            # stop worker application
            self._assigner_ref.remove_apply(self._op_key)
        self.state = OperandState.CANCELLED
        self.ref().start_operand(_tell=True)

    @log_unhandled
    def _handle_cancelled(self):
        if self._worker_endpoint:
            self._resource_ref.deallocate_resource(self._session_id, self._op_key, self._worker_endpoint)
        if self._is_terminal:
            self._graph_ref.mark_terminal_finished(self._op_key, final_state=GraphState.CANCELLED, _tell=True)
        for k in self._succ_keys:
            self._get_operator_actor(k).propagate_state(OperandState.CANCELLING, _tell=True)

    @log_unhandled
    def _handle_freed(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2017 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import HasInput
from .. import opcodes as OperandDef
from ..serialize import AnyField, KeyField, TupleField, \
    DataTypeField, Int32Field, Int64Field, BoolField


class Reduction(HasInput):
    _input = KeyField('input')
    _out = KeyField('out')
    _axis = AnyField('axis')  # can be None or int or tuple of ints, just infer the data
    _dtype = DataTypeField('dtype')
    _keepdims = BoolField('keepdims')
    _combine_size = Int32Field('combine_size')

    @property
    def axis(self):
        return getattr(self, '_axis', None)

    @property
    def dtype(self):
        return getattr(self, '_dtype', None)

    @property
    def keepdims(self):
        return getattr(self, '_keepdims', None)

    @property
    def combine_size(self):
        return getattr(self, '_combine_size', None)


class Sum(Reduction):
    _op_type_ = OperandDef.SUM


class NanSum(Reduction):
    _op_type_ = OperandDef.NANSUM


class Prod(Reduction):
    _op_type_ = OperandDef.PROD


class NanProd(Reduction):
    _op_type_ = OperandDef.NANPROD


class Max(Reduction):
    _op_type_ = OperandDef.MAX


class NanMax(Reduction):
    _op_type_ = OperandDef.NANMAX


class Min(Reduction):
    _op_type_ = OperandDef.MIN


class NanMin(Reduction):
    _op_type_ = OperandDef.NANMIN


class All(Reduction):
    _op_type_ = OperandDef.ALL


class Any(Reduction):
    _op_type_ = OperandDef.ANY


class MeanChunk(Reduction):
    _op_type_ = OperandDef.MEAN_CHUNK


class MeanCombine(Reduction):
    _op_type_ = OperandDef.MEAN_COMBINE


class Mean(Reduction):
    _op_type_ = OperandDef.MEAN


class NanMeanChunk(Reduction):
    _op_type_ = OperandDef.NANMEAN_CHUNK


class NanMean(Reduction):
    _op_type_ = OperandDef.NANMEAN


class Argmax(Reduction):
    _op_type_ = OperandDef.ARGMAX


class ArgmaxChunk(Reduction):
    _op_type_ = OperandDef.ARGMAX_CHUNK

    _offset = Int64Field('offset')
    _total_shape = TupleField('total_shape')

    @property
    def offset(self):
        return getattr(self, '_offset', None)

    @property
    def total_shape(self):
        return getattr(self, '_total_shape', None)


class ArgmaxCombine(Reduction):
    _op_type_ = OperandDef.ARGMAX_COMBINE


class NanArgmax(Reduction):
    _op_type_ = OperandDef.NANARGMAX


class NanArgmaxChunk(Reduction):
    _op_type_ = OperandDef.NANARGMAX_CHUNK

    _offset = Int64Field('offset')
    _total_shape = TupleField('total_shape')

    @property
    def offset(self):
        return getattr(self, '_offset', None)

    @property
    def total_shape(self):
        return getattr(self, '_total_shape', None)


class NanArgmaxCombine(Reduction):
    _op_type_ = OperandDef.NANARGMAX_COMBINE


class Argmin(Reduction):
    _op_type_ = OperandDef.ARGMIN


class ArgminChunk(Reduction):
    _op_type_ = OperandDef.ARGMIN_CHUNK

    _offset = Int64Field('offset')
    _total_shape = TupleField('total_shape')

    @property
    def offset(self):
        return getattr(self, '_offset', None)

    @property
    def total_shape(self):
        return getattr(self, '_total_shape', None)


class ArgminCombine(Reduction):
    _op_type_ = OperandDef.ARGMIN_COMBINE


class NanArgmin(Reduction):
    _op_type_ = OperandDef.NANARGMIN


class NanArgminChunk(Reduction):
    _op_type_ = OperandDef.NANARGMIN_CHUNK

    _offset = Int64Field('offset')
    _total_shape = TupleField('total_shape')

    @property
    def offset(self):
        return getattr(self, '_offset', None)

    @property
    def total_shape(self):
        return getattr(self, '_total_shape', None)


class NanArgminCombine(Reduction):
    _op_type_ = OperandDef.NANARGMIN_COMBINE


class CountNonzero(Reduction):
    _op_type_ = OperandDef.COUNT_NONZERO


class MomentChunk(Reduction):
    _op_type_ = OperandDef.MOMENT_CHUNK

    _moment = Int32Field('moment', default=2)

    @property
    def moment(self):
        return getattr(self, '_moment', 2)


class MomentCombine(Reduction):
    _op_type_ = OperandDef.MOMENT_COMBINE

    _moment = Int32Field('moment')

    @property
    def moment(self):
        return getattr(self, '_moment', 2)


class Moment(Reduction):
    _op_type_ = OperandDef.MOMENT

    _moment = Int32Field('moment', default=2)
    _ddof = Int32Field('ddof')

    @property
    def moment(self):
        return getattr(self, '_moment', 2)

    @property
    def ddof(self):
        return self._ddof


class Var(Reduction):
    _op_type_ = OperandDef.VAR

    _ddof = Int32Field('ddof')

    @property
    def ddof(self):
        return self._ddof


class NanVar(Reduction):
    _op_type_ = OperandDef.NANVAR

    _ddof = Int32Field('ddof')

    @property
    def ddof(self):
        return self._ddof


class NanMomentChunk(Reduction):
    _op_type_ = OperandDef.NANMOMENT_CHUNK

    _moment = Int32Field('moment', default=2)

    @property
    def moment(self):
        return getattr(self, '_moment', 2)


class NanMomentCombine(Reduction):
    _op_type_ = OperandDef.NANMOMENT_COMBINE

    _moment = Int32Field('moment')

    @property
    def moment(self):
        return getattr(self, '_moment', 2)


class NanMoment(Reduction):
    _op_type_ = OperandDef.NANMOMENT

    _moment = Int32Field('moment', default=2)
    _ddof = Int32Field('ddof')

    @property
    def moment(self):
        return getattr(self, '_moment', 2)

    @property
    def ddof(self):
        return self._ddof


class CumReduction(HasInput):
    _input = KeyField('input')
    _axis = Int32Field('axis')
    _out = KeyField('out')

    @property
    def axis(self):
        return getattr(self, '_axis', None)

    @property
    def dtype(self):
        return getattr(self, '_dtype', None)


class Cumsum(CumReduction):
    _op_type_ = OperandDef.CUMSUM


class Cumprod(CumReduction):
    _op_type_ = OperandDef.CUMPROD


class NanCumsum(CumReduction):
    _op_type_ = OperandDef.NANCUMSUM


class NanCumprod(CumReduction):
    _op_type_ = OperandDef.NANCUMPROD



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import struct
import sys
import threading
import weakref

import numpy as np

from .compat import getargspec, six
from .actors import FunctionActor
from .actors.core import ActorRef
from .errors import PromiseTimeout
from .utils import wraps

logger = logging.getLogger(__name__)
_promise_pool = dict()


class Promise(object):
    """
    Object representing a promised result
    """
    def __init__(self, resolve=None, reject=None, done=False):
        # use random promise id
        self._id = struct.pack('<Q', id(self)) + np.random.bytes(32)
        # register in global pool to reject gc collection
        _promise_pool[self._id] = self

        self._accept_handler = self._wrap_handler(resolve)
        self._reject_handler = self._wrap_handler(reject)

        # _bind_item indicates the Promise object whose step_next()
        # should be called when current Promise finishes. For instance,
        # in ``p2 = p1.then(lambda arg: promise_call())``, the value of
        # _bind_item of the Promise object returned by ``promise_call()``
        # is p2, thus when the lambda finishes, subsequent operations in
        # p2 can be executed.
        self._bind_item = None  # type: Promise
        # _next_item indicates the next Promise object to be invoked
        # when the current one returns.
        self._next_item = None  # type: Promise

        # promise results
        self._accepted = None if not done else True
        self._args = ()

        self.post_create()

    def __del__(self):
        self.pre_destroy()

    @property
    def id(self):
        return self._id

    def post_create(self):
        pass

    def pre_destroy(self):
        pass

    def _wrap_handler(self, func):
        """
        Wraps a promise handler
        """
        if func is None:
            return None

        @wraps(func)
        def _wrapped(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
                if isinstance(result, Promise):
                    # the function itself returns a promise object
                    # bind returned promise object to current promise
                    result._bind_item = self
                    if result._accepted is not None:
                        # promise already done, we move next
                        result.step_next(_accept=result._accepted)
                else:
                    # return non-promise result, we just step next
                    self.step_next(result)
            except:
                # error occurred when executing func, we reject with exc_info
                logger.exception('Exception met in executing promise.')
                exc = sys.exc_info()
                self.step_next(*exc, _accept=False)
        return _wrapped

    def _get_bind_root(self):
        """
        Get root promise of result promises
        :return: root promise
        :rtype: Promise
        """
        target = self
        while target._bind_item is not None:
            if target and target.id in _promise_pool:
                # remove binder that will not be used later
                del _promise_pool[target.id]
            target = target._bind_item
        return target

    @staticmethod
    def _get_handling_promise(p, handler_attr):
        """
        Get a promise object that defines the handler needed
        :param Promise p: promise object
        :param str handler_attr: attribute name of the handler
        :rtype: Promise
        """
        while getattr(p, handler_attr) is None:
            p = p._get_bind_root()
            if p and p._next_item is not None:
                if p.id in _promise_pool:
                    # remove predecessor that will not be used later
                    del _promise_pool[p.id]
                p = p._next_item
            else:
                break
        return p

    def step_next(self, *args, **kwargs):
        """
        Step into next promise with given args and kwargs
        """
        def _log_unexpected_error():
            if args and len(args) == 3 and issubclass(args[0], Exception):
                try:
                    six.reraise(*args)
                except:
                    logger.exception('Unhandled exception in promise')

        accept = kwargs.pop('_accept', True)
        target_promise = self

        self._accepted = accept
        if not self._accepted:
            # we only preserve exceptions to avoid tracing huge objects
            self._args = args

        try:
            target_promise = self._get_bind_root()

            if target_promise and target_promise.id in _promise_pool:
                del _promise_pool[target_promise.id]

            target_promise = target_promise._next_item
            if not target_promise:
                if not accept:
                    _log_unexpected_error()
                return

            if accept:
                acceptor = self._get_handling_promise(target_promise, '_accept_handler')
                if acceptor and acceptor._accept_handler:
                    acceptor._accept_handler(*args, **kwargs)
            else:
                rejecter = self._get_handling_promise(target_promise, '_reject_handler')
                if rejecter and rejecter._reject_handler:
                    rejecter._reject_handler(*args, **kwargs)
                else:
                    _log_unexpected_error()
        finally:
            if target_promise and target_promise.id in _promise_pool:
                del _promise_pool[target_promise.id]

    def then(self, on_fulfilled, on_rejected=None):
        promise = Promise(on_fulfilled, on_rejected)
        self._next_item = promise
        if self._accepted is not None:
            self.step_next(*self._args, **dict(_accept=self._accepted))
        return promise

    def catch(self, on_rejected):
        return self.then(None, on_rejected)

    def wait(self, waiter=None, timeout=None):
        """
        Wait when the promise returns. Currently only used in debug.
        :param waiter: wait object
        :param timeout: wait timeout
        :return: accept or reject
        """
        waiter = threading.Event()
        status = []

        def _finish_exec(accept_or_reject):
            waiter.set()
            status.append(accept_or_reject)

        self.then(lambda *_, **__: _finish_exec(True),
                  lambda *_, **__: _finish_exec(False))
        waiter.wait(timeout)
        return status[0]


class PromiseRefWrapper(object):
    """
    Promise wrapper that enables promise call by adding _promise=True
    """
    def __init__(self, ref, caller):
        self._ref = ref
        self._caller = caller  # type: PromiseActor

    def send(self, message):
        return self._ref.send(self, message)

    def tell(self, message, delay=None):
        return self._ref.tell(self, message, delay=delay)

    def destroy(self):
        return self._ref.destroy_actor(self)

    @property
    def uid(self):
        return self._ref.uid

    @property
    def ctx(self):
        return self._ref.ctx

    @property
    def address(self):
        return self._ref.address

    def __getattr__(self, item):
        if item.startswith('_'):
            return object.__getattribute__(self, item)

        def _mt_fun(*args, **kwargs):
            ref_fun = getattr(self._ref, item)
            if not kwargs.pop('_promise', False):
                return ref_fun(*args, **kwargs)

            p = Promise()
            self._caller._promises[p.id] = p

            timeout = kwargs.pop('_timeout', 0)

            kwargs['callback'] = ((self._caller.uid, self._caller.address),
                                  'handle_promise', p.id)
            kwargs['_tell'] = True
            ref_fun(*args, **kwargs)

            if timeout > 0:
                # add a callback that triggers some times later to deal with timeout
                self._caller.ref().handle_promise_timeout(p.id, _tell=True, _delay=timeout)

            return p

        return _mt_fun


def reject_on_exception(func):
    """
    Decorator on actor callback functions that handles exceptions by
    sending it to caller as promise rejections. The function should have
    an argument called ``callback``.
    """
    arg_names = getargspec(func).args
    callback_pos = None
    if arg_names:
        for idx, name in enumerate(arg_names):
            if name == 'callback':
                callback_pos = idx
                break

    @wraps(func)
    def _wrapped(*args, **kwargs):
        callback = None
        if 'callback' in kwargs:
            callback = kwargs['callback']
        elif callback_pos and callback_pos < len(args):
            callback = args[callback_pos]

        try:
            return func(*args, **kwargs)
        except:
            actor = args[0]
            logger.exception('Unhandled exception in promise call')
            if callback:
                actor.tell_promise(callback, *sys.exc_info(), **dict(_accept=False))
    return _wrapped


class PromiseActor(FunctionActor):
    """
    Actor class providing promise functionality
    """
    def promise_ref(self, *args, **kwargs):
        """
        Wraps an existing ActorRef into a promise ref
        """
        if not hasattr(self, '_promises'):
            self._promises = weakref.WeakValueDictionary()  # type: dict[object, Promise]

        if not args and not kwargs:
            ref = self.ref()
        elif args and isinstance(args[0], ActorRef):
            ref = self.ctx.actor_ref(args[0].uid, address=args[0].address)
        else:
            ref = self.ctx.actor_ref(*args, **kwargs)
        return PromiseRefWrapper(ref, self)

    def tell_promise(self, callback, *args, **kwargs):
        """
        Tell promise results to the caller
        :param callback: promise callback
        """
        uid, address = callback[0]
        callback_args = callback[1:] + args + (kwargs, )
        self.ctx.actor_ref(uid, address=address).tell(callback_args)

    def handle_promise(self, promise_id, *args, **kwargs):
        """
        Callback entry for promise results
        :param promise_id: promise key
        """
        try:
            self._promises[promise_id].step_next(*args, **kwargs)
            if promise_id in self._promises:
                del self._promises[promise_id]
        except KeyError:
            logger.warning('Promise %r reentered in %s', promise_id, self.uid)

    def handle_promise_timeout(self, promise_id):
        """
        Callback entry for promise timeout
        :param promise_id: promise key
        """
        p = self._promises.get(promise_id)
        if not p or p._accepted is not None:
            # skip promises that are already finished
            return

        try:
            del self._promises[promise_id]
        except KeyError:
            pass

        try:
            raise PromiseTimeout
        except PromiseTimeout:
            exc_info = sys.exc_info()
        p.step_next(*exc_info, **dict(_accept=False))


def all_(promises):
    """
    Create a promise with promises. Invoked when all referenced promises accepted
    or at least one rejected
    :param promises: collection of promises
    :return: the new promise
    """
    promises = list(promises)
    new_promise = Promise()
    finish_set = set()

    def _build_then(promise):
        def _then(*_, **__):
            finish_set.add(promise.id)
            if all(p.id in finish_set for p in promises):
                new_promise.step_next()
        return _then

    def _handle_reject(*args, **kw):
        if new_promise._accepted is not None:
            return
        kw['_accept'] = False
        new_promise.step_next(*args, **kw)

    [p.then(_build_then(p), _handle_reject) for p in promises if isinstance(p, Promise)]
    if promises:
        return new_promise
    else:
        new_promise.step_next()
        return new_promise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import json
import time
import numpy as np

from .api import MarsAPI
from .scheduler.graph import GraphState
from .serialize import dataserializer


class LocalSession(object):
    def __init__(self):
        from .tensor.execution.core import Executor

        self._executor = Executor()

    def run(self, *tensors, **kw):
        if self._executor is None:
            raise RuntimeError('Session has closed')
        return self._executor.execute_tensors(tensors, **kw)

    def decref(self, *keys):
        self._executor.decref(*keys)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self._executor = None


class Session(object):
    _default_session = None

    def __init__(self, endpoint=None):
        if endpoint is not None:
            if 'http' in endpoint:
                # connect to web
                from .web.session import Session as WebSession

                self._sess = WebSession(endpoint)
            else:
                # connect to local cluster
                from .deploy.local.session import LocalClusterSession

                self._sess = LocalClusterSession(endpoint)
        else:
            self._sess = LocalSession()

        self._executed_keys = set()

    def run(self, *tensors, **kw):
        from . import tensor as mt

        ret_list = False
        if len(tensors) == 1 and isinstance(tensors[0], (tuple, list)):
            ret_list = True
            tensors = tensors[0]
        elif len(tensors) > 1:
            ret_list = True

        tensors = tuple(mt.tensor(t) for t in tensors)
        result = self._sess.run(*tensors, **kw)
        self._executed_keys.update(t.key for t in tensors)
        for t in tensors:
            t._execute_session = self

        ret = []
        for r, t in zip(result, tensors):
            if r is None:
                ret.append(r)
                continue
            if t.isscalar() and hasattr(r, 'item'):
                ret.append(np.asscalar(r))
            else:
                ret.append(r)
        if ret_list:
            return ret
        return ret[0]

    def decref(self, *keys):
        if hasattr(self._sess, 'decref'):
            self._sess.decref(*keys)

    def __getattr__(self, attr):
        try:
            obj = self._sess.__getattribute__(attr)
            return obj
        except AttributeError:
            raise

    def __enter__(self):
        self._sess.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._sess.__exit__(exc_type, exc_val, exc_tb)

    close = __exit__

    def as_default(self):
        Session._default_session = self
        return self

    @classmethod
    def default_or_local(cls):
        if cls._default_session is not None:
            return cls._default_session

        cls._default_session = Session()
        return cls._default_session


def new_session(scheduler=None):
    return Session(scheduler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def tile(A, reps):
    """
    Construct a tensor by repeating A the number of times given by reps.

    If `reps` has length ``d``, the result will have dimension of
    ``max(d, A.ndim)``.

    If ``A.ndim < d``, `A` is promoted to be d-dimensional by prepending new
    axes. So a shape (3,) array is promoted to (1, 3) for 2-D replication,
    or shape (1, 1, 3) for 3-D replication. If this is not the desired
    behavior, promote `A` to d-dimensions manually before calling this
    function.

    If ``A.ndim > d``, `reps` is promoted to `A`.ndim by pre-pending 1's to it.
    Thus for an `A` of shape (2, 3, 4, 5), a `reps` of (2, 2) is treated as
    (1, 1, 2, 2).

    Note : Although tile may be used for broadcasting, it is strongly
    recommended to use Mars' broadcasting operations and functions.

    Parameters
    ----------
    A : array_like
        The input tensor.
    reps : array_like
        The number of repetitions of `A` along each axis.

    Returns
    -------
    c : Tensor
        The tiled output tensor.

    See Also
    --------
    repeat : Repeat elements of a tensor.
    broadcast_to : Broadcast a tensor to a new shape

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([0, 1, 2])
    >>> mt.tile(a, 2).execute()
    array([0, 1, 2, 0, 1, 2])
    >>> mt.tile(a, (2, 2)).execute()
    array([[0, 1, 2, 0, 1, 2],
           [0, 1, 2, 0, 1, 2]])
    >>> mt.tile(a, (2, 1, 2)).execute()
    array([[[0, 1, 2, 0, 1, 2]],
           [[0, 1, 2, 0, 1, 2]]])

    >>> b = mt.array([[1, 2], [3, 4]])
    >>> mt.tile(b, 2).execute()
    array([[1, 2, 1, 2],
           [3, 4, 3, 4]])
    >>> mt.tile(b, (2, 1)).execute()
    array([[1, 2],
           [3, 4],
           [1, 2],
           [3, 4]])

    >>> c = mt.array([1,2,3,4])
    >>> mt.tile(c,(4,1)).execute()
    array([[1, 2, 3, 4],
           [1, 2, 3, 4],
           [1, 2, 3, 4],
           [1, 2, 3, 4]])
    """
    from ..merge import concatenate

    try:
        tup = tuple(reps)
    except TypeError:
        tup = (reps,)

    d = len(tup)
    if A.ndim < d:
        A = A[tuple(np.newaxis for _ in range(d - A.ndim))]
    elif A.ndim > d:
        tup = (1,) * (A.ndim - d) + tup

    a = A
    for axis, rep in enumerate(tup):
        if rep == 0:
            slc = (slice(None),) * axis + (slice(0),)
            a = a[slc]
        elif rep < 0:
            raise ValueError('negative dimensions are not allowed')
        elif rep > 1:
            a = concatenate([a] * rep, axis=axis)

    return a



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin


class TensorRandint(operands.Randint, TensorRandomOperandMixin):
    __slots__ = '_low', '_high', '_density', '_size'

    def __init__(self, state=None, size=None, dtype=None,
                 low=None, high=None, sparse=False, density=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorRandint, self).__init__(_state=state, _size=size,
                                            _low=low, _high=high, _dtype=dtype,
                                            _sparse=sparse, _density=density,
                                            _gpu=gpu, **kw)

    def __call__(self, chunks=None):
        return self.new_tensor(None, None, raw_chunks=chunks)


def randint(random_state, low, high=None, size=None, dtype='l', density=None,
            chunks=None, gpu=None, **kw):
    """
    Return random integers from `low` (inclusive) to `high` (exclusive).

    Return random integers from the "discrete uniform" distribution of
    the specified dtype in the "half-open" interval [`low`, `high`). If
    `high` is None (the default), then results are from [0, `low`).

    Parameters
    ----------
    low : int
        Lowest (signed) integer to be drawn from the distribution (unless
        ``high=None``, in which case this parameter is one above the
        *highest* such integer).
    high : int, optional
        If provided, one above the largest (signed) integer to be drawn
        from the distribution (see above for behavior if ``high=None``).
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    dtype : dtype, optional
        Desired dtype of the result. All dtypes are determined by their
        name, i.e., 'int64', 'int', etc, so byteorder is not available
        and a specific precision may have different C types depending
        on the platform. The default value is 'np.int'.
    density: float, optional
        if density specified, a sparse tensor will be created
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : int or Tensor of ints
        `size`-shaped tensor of random integers from the appropriate
        distribution, or a single such random int if `size` not provided.

    See Also
    --------
    random.random_integers : similar to `randint`, only for the closed
        interval [`low`, `high`], and 1 is the lowest value if `high` is
        omitted. In particular, this other one is the one to use to generate
        uniformly distributed discrete non-integers.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.random.randint(2, size=10).execute()
    array([1, 0, 0, 0, 1, 1, 0, 0, 1, 0])
    >>> mt.random.randint(1, size=10).execute()
    array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    Generate a 2 x 4 tensor of ints between 0 and 4, inclusive:

    >>> mt.random.randint(5, size=(2, 4)).execute()
    array([[4, 0, 2, 1],
           [3, 2, 2, 0]])
    """
    sparse = bool(density)
    size = random_state._handle_size(size)
    op = TensorRandint(state=random_state._state, low=low, high=high, size=size, dtype=dtype,
                       gpu=gpu, sparse=sparse, density=density, **kw)
    return op(chunks=chunks)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numexpr as ne
import numpy as np

from ...compat import six, izip
from ..expressions import arithmetic, reduction
from ..expressions.fuse import TensorNeFuseChunk
from .array import as_same_device

_VAR_FLAG = 'V_'

NE_UNARYOP_TO_STRING = {
    arithmetic.TensorNegative: '-',
    arithmetic.TensorAbs: 'abs',
    arithmetic.TensorConj: 'conj',
    arithmetic.TensorExp: 'exp',
    arithmetic.TensorLog: 'log',
    arithmetic.TensorLog10: 'log10',
    arithmetic.TensorExpm1: 'expm1',
    arithmetic.TensorLog1p: 'log1p',
    arithmetic.TensorSqrt: 'sqrt',

    arithmetic.TensorSin: 'sin',
    arithmetic.TensorCos: 'cos',
    arithmetic.TensorTan: 'tan',
    arithmetic.TensorArcsin: 'arcsin',
    arithmetic.TensorArccos: 'arccos',
    arithmetic.TensorArctan: 'arctan',
    arithmetic.TensorSinh: 'sinh',
    arithmetic.TensorCosh: 'cosh',
    arithmetic.TensorTanh: 'tanh',
    arithmetic.TensorArcsinh: 'arcsinh',
    arithmetic.TensorArccosh: 'arccosh',
    arithmetic.TensorArctanh: 'arctanh'
}


NE_BINOP_TO_STRING = {
    arithmetic.TensorAdd: '+',
    arithmetic.TensorAddConstant: '+',
    arithmetic.TensorTreeAdd: '+',
    arithmetic.TensorSubtract: '-',
    arithmetic.TensorSubConstant: '-',
    arithmetic.TensorMultiply: '*',
    arithmetic.TensorMulConstant: '*',
    arithmetic.TensorTreeMultiply: '*',
    arithmetic.TensorDivide: '/',
    arithmetic.TensorDivConstant: '/',
    arithmetic.TensorMod: '%',
    arithmetic.TensorModConstant: '%',
    arithmetic.TensorPower: '**',
    arithmetic.TensorPowConstant: '**',
    arithmetic.TensorLshift: '<<',
    arithmetic.TensorLshiftConstant: '<<',
    arithmetic.TensorRshift: '>>',
    arithmetic.TensorRshiftConstant: '>>',

    arithmetic.TensorEqual: '==',
    arithmetic.TensorEqConstant: '==',
    arithmetic.TensorNotEqual: '!=',
    arithmetic.TensorNeConstant: '!=',
    arithmetic.TensorLessThan: '<',
    arithmetic.TensorLtConstant: '<',
    arithmetic.TensorLessEqual: '<=',
    arithmetic.TensorLeConstant: '<=',
    arithmetic.TensorGreaterThan: '>',
    arithmetic.TensorGtConstant: '>',
    arithmetic.TensorGreaterEqual: '>=',
    arithmetic.TensorGeConstant: '>='
}


NE_REDUCTION_TO_STRING = {
    reduction.TensorSum: 'sum',
    reduction.TensorProd: 'prod',
    reduction.TensorMax: 'max',
    reduction.TensorMin: 'min'
}


def _handle_unary(chunk):
    if len(chunk.inputs) != 1:
        raise ValueError('unary operand inputs should be 1')
    data = chunk.inputs[0]
    unary_op = NE_UNARYOP_TO_STRING[type(chunk.op)]
    _expr = '{}({})'.format(unary_op, _VAR_FLAG + data.key)
    return _expr


def _decompose(chunk):
    expr = _VAR_FLAG + chunk.key
    for node, op in zip(reversed(chunk.composed), reversed(chunk.op.operands)):
        _expr = _evalute(node)
        expr = expr.replace(_VAR_FLAG + node.key, '({})'.format(_expr))
    return expr


def _handle_bin(chunk):
    const = getattr(chunk.op, 'constant', [])
    reverse = getattr(chunk.op, 'reverse', False)
    op = NE_BINOP_TO_STRING[type(chunk.op)]

    exprs = [op.join(_VAR_FLAG + c.key for c in chunk.inputs), op.join(str(c) for c in const)]

    if reverse:
        exprs = list(reversed(exprs))

    return op.join(filter(None, exprs))


def _wrap_bool(data):
    if data.dtype == np.bool_:
        return lambda x: 'where({0}, 1, 0)'.format(x)

    return lambda x: x


def _handle_reduction(chunk):
    ax = chunk.op.axis
    data = chunk.inputs[0]
    op_str = NE_REDUCTION_TO_STRING[type(chunk.op)]
    # TODO(hks): delete it if numexpr.sum fixed
    if len(ax) == data.ndim:
        _expr = '{}({})'.format(op_str, _wrap_bool(data)(_VAR_FLAG + data.key))
    elif len(ax) == 1:
        if data.shape[ax[0]] == 1:
            _expr = _VAR_FLAG + data.key
        else:
            _expr = '{}({},axis={})'.format(op_str, _wrap_bool(data)(_VAR_FLAG + data.key), ax[0])
    else:
        raise ValueError("numexpr cannot encode axis")
    return _expr


def _evalute(chunk):
    if type(chunk.op) in NE_UNARYOP_TO_STRING:
        return _handle_unary(chunk)
    elif type(chunk.op) in NE_BINOP_TO_STRING:
        return _handle_bin(chunk)
    elif type(chunk.op) in NE_REDUCTION_TO_STRING:
        return _handle_reduction(chunk)
    elif type(chunk.op) == TensorNeFuseChunk:
        return _decompose(chunk)
    else:
        raise TypeError("unsupported operator in numexpr: {}".format(chunk.op.__class__.__name__))


def _maybe_keepdims(chunk, res):
    out_chunk = chunk.composed[-1] if type(chunk.op) == TensorNeFuseChunk else chunk
    if type(out_chunk.op) in NE_REDUCTION_TO_STRING and out_chunk.op.keepdims:
        res = np.reshape(res, out_chunk.shape)
    return res


def evaluate(ctx, chunk):
    inputs = as_same_device([ctx[c.key] for c in chunk.inputs], device=chunk.op.device)
    for c, i in izip(chunk.inputs, inputs):
        six.exec_('V_' + c.key + ' = i')
    expr = _evalute(chunk)
    res = ne.evaluate(expr)
    res = _maybe_keepdims(chunk, res)
    ctx[chunk.key] = res


def register_numexpr_handler():
    from .core import register

    register(TensorNeFuseChunk, evaluate)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import inject_dtype
from .core import TensorUnaryOp


class TensorIsNan(operands.IsNan, TensorUnaryOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorIsNan, self).__init__(_casting=casting, _err=err,
                                          _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x):
        return x.issparse()


@inject_dtype(np.bool_)
def isnan(x, out=None, where=None, **kwargs):
    """
    Test element-wise for NaN and return result as a boolean tensor.

    Parameters
    ----------
    x : array_like
        Input tensor.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor or bool
        For scalar input, the result is a new boolean with value True if
        the input is NaN; otherwise the value is False.

        For array input, the result is a boolean tensor of the same
        dimensions as the input and the values are True if the
        corresponding element of the input is NaN; otherwise the values are
        False.

    See Also
    --------
    isinf, isneginf, isposinf, isfinite, isnat

    Notes
    -----
    Mars uses the IEEE Standard for Binary Floating-Point for Arithmetic
    (IEEE 754). This means that Not a Number is not equivalent to infinity.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.isnan(mt.nan).execute()
    True
    >>> mt.isnan(mt.inf).execute()
    False
    >>> mt.isnan([mt.log(-1.).execute(),1.,mt.log(0).execute()]).execute()
    array([ True, False, False])
    """
    op = TensorIsNan(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..datasource import tensor as astensor


def atleast_2d(*tensors):
    """
    View inputs as tensors with at least two dimensions.

    Parameters
    ----------
    tensors1, tensors2, ... : array_like
        One or more array-like sequences.  Non-tensor inputs are converted
        to tensors.  Tensors that already have two or more dimensions are
        preserved.

    Returns
    -------
    res, res2, ... : Tensor
        A tensor, or list of tensors, each with ``a.ndim >= 2``.
        Copies are avoided where possible, and views with two or more
        dimensions are returned.

    See Also
    --------
    atleast_1d, atleast_3d

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.session import new_session

    >>> sess = new_session().as_default()

    >>> mt.atleast_2d(3.0).execute()
    array([[ 3.]])

    >>> x = mt.arange(3.0)
    >>> mt.atleast_2d(x).execute()
    array([[ 0.,  1.,  2.]])

    >>> sess.run(mt.atleast_2d(1, [1, 2], [[1, 2]]))
    [array([[1]]), array([[1, 2]]), array([[1, 2]])]

    """
    new_tensors = []
    for x in tensors:
        x = astensor(x)
        if x.ndim == 0:
            x = x[np.newaxis, np.newaxis]
        elif x.ndim == 1:
            x = x[np.newaxis, :]

        new_tensors.append(x)

    if len(new_tensors) == 1:
        return new_tensors[0]
    return new_tensors



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import threading
import logging
import functools
from collections import defaultdict

from bokeh.application import Application
from bokeh.application.handlers import FunctionHandler
from bokeh.server.server import Server
import jinja2
from tornado import web, ioloop

from .. import kvstore
from ..compat import six
from ..utils import get_next_port
from ..config import options
from ..scheduler import GraphActor, ResourceActor
from ..api import MarsAPI

logger = logging.getLogger(__name__)


def get_jinja_env():
    from datetime import datetime
    from ..utils import readable_size

    _jinja_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(os.path.join(os.path.dirname(__file__), 'templates')),
        )

    def format_ts(value):
        if value is None:
            return None
        return datetime.fromtimestamp(value).strftime('%Y-%m-%d %H:%M:%S')

    _jinja_env.filters['format_ts'] = format_ts
    _jinja_env.filters['readable_size'] = readable_size
    return _jinja_env


class BokehStaticFileHandler(web.StaticFileHandler):
    @classmethod
    def get_absolute_path(cls, root, path):
        from bokeh import server
        path_parts = path.rsplit('/', 1)
        if 'bokeh' in path_parts[-1]:
            root = os.path.join(os.path.dirname(server.__file__), "static")
        return super(BokehStaticFileHandler, cls).get_absolute_path(root, path)

    def validate_absolute_path(self, root, absolute_path):
        from bokeh import server
        path_parts = absolute_path.rsplit('/', 1)
        if 'bokeh' in path_parts[-1]:
            root = os.path.join(os.path.dirname(server.__file__), "static")
        return super(BokehStaticFileHandler, self).validate_absolute_path(root, absolute_path)


class MarsWebAPI(MarsAPI):
    def __init__(self, scheduler_ip):
        super(MarsWebAPI, self).__init__(scheduler_ip)

    def get_tasks_info(self):
        sessions = defaultdict(dict)
        for session_id, session_ref in six.iteritems(self.session_manager.get_sessions()):
            session_desc = sessions[session_id]
            session_desc['id'] = session_id
            session_desc['name'] = session_id
            session_desc['tasks'] = dict()
            session_ref = self.actor_client.actor_ref(session_ref)
            for graph_key, graph_ref in six.iteritems(session_ref.get_graph_refs()):
                task_desc = dict()

                state = self.kv_store.read(
                    '/sessions/%s/graph/%s/state' % (session_id, graph_key)).value
                if state == 'PREPARING':
                    task_desc['state'] = state.lower()
                    session_desc['tasks'][graph_key] = task_desc
                    continue

                graph_ref = self.actor_client.actor_ref(graph_ref)
                task_desc['id'] = graph_key
                task_desc['state'] = graph_ref.get_state().value
                start_time, end_time, graph_size = graph_ref.get_graph_info()
                task_desc['start_time'] = start_time
                task_desc['end_time'] = end_time or 'N/A'
                task_desc['graph_size'] = graph_size or 'N/A'

                session_desc['tasks'][graph_key] = task_desc
        return sessions

    def get_task_detail(self, session_id, task_id):
        graph_uid = GraphActor.gen_name(session_id, task_id)
        graph_ref = self.get_actor_ref(graph_uid)
        return graph_ref.calc_stats()

    def get_workers_meta(self):
        resource_uid = ResourceActor.default_name()
        resource_ref = self.get_actor_ref(resource_uid)
        return resource_ref.get_workers_meta()


class MarsWeb(object):
    def __init__(self, port=None, scheduler_ip=None):
        self._port = port
        self._scheduler_ip = scheduler_ip
        self._server = None
        self._server_thread = None

    @property
    def port(self):
        return self._port

    def start(self, event=None, block=False):
        try:
            ioloop.IOLoop.current()
        except RuntimeError:
            if six.PY3:
                import asyncio
                asyncio.set_event_loop(asyncio.new_event_loop())
                loop = None
                try:
                    loop = ioloop.IOLoop.current()
                except:
                    pass
                if loop is None:
                    raise
            else:
                raise

        if self._scheduler_ip is None:
            kv_store = kvstore.get(options.kv_store)
            try:
                schedulers = [s.key.rsplit('/', 1)[1] for s in kv_store.read('/schedulers').children]
                self._scheduler_ip = schedulers[0]
            except KeyError:
                raise KeyError('No scheduler is available')

        static_path = os.path.join(os.path.dirname(__file__), 'static')

        handlers = dict()
        for p, h in _ui_handlers.items():
            handlers[p] = Application(FunctionHandler(functools.partial(h, self._scheduler_ip)))
        extra_patterns = [
            ('/static/(.*)', BokehStaticFileHandler, {'path': static_path})
        ]
        for p, h in _api_handlers.items():
            extra_patterns.append((p, h, {'scheduler_ip': self._scheduler_ip}))

        retrial = 5
        while retrial:
            try:
                if self._port is None:
                    use_port = get_next_port()
                else:
                    use_port = self._port
                self._server = Server(
                    handlers, allow_websocket_origin=['*'],
                    address='0.0.0.0', port=use_port,
                    extra_patterns=extra_patterns,
                )
                self._server.start()
                self._port = use_port
                logger.info('Mars UI started at 0.0.0.0:%d', self._port)
                break
            except:
                if self._port is not None:
                    raise
                retrial -= 1
                if retrial == 0:
                    raise

        if not block:
            self._server_thread = threading.Thread(target=self._server.io_loop.start)
            self._server_thread.daemon = True
            self._server_thread.start()

            if event:
                event.set()
        else:
            if event:
                event.set()

            self._server.io_loop.start()

    def stop(self):
        if self._server is not None:
            self._server.io_loop.stop()
            self._server.stop()


_ui_handlers = dict()
_api_handlers = dict()


def register_ui_handler(pattern, handler):
    _ui_handlers[pattern] = handler


def register_api_handler(pattern, handler):
    _api_handlers[pattern] = handler



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import inject_dtype
from .core import TensorUnaryOp


class TensorIsReal(operands.IsReal, TensorUnaryOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorIsReal, self).__init__(_casting=casting, _err=err,
                                           _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x):
        return False


@inject_dtype(np.bool_)
def isreal(x, **kwargs):
    """
    Returns a bool tensor, where True if input element is real.

    If element has complex type with zero complex part, the return value
    for that element is True.

    Parameters
    ----------
    x : array_like
        Input tensor.

    Returns
    -------
    out : Tensor, bool
        Boolean tensor of same shape as `x`.

    See Also
    --------
    iscomplex
    isrealobj : Return True if x is not a complex type.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.isreal([1+1j, 1+0j, 4.5, 3, 2, 2j]).execute()
    array([False,  True,  True,  True,  True, False])

    """
    op = TensorIsReal(**kwargs)
    return op(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorPoisson(operands.Poisson, TensorRandomOperandMixin):
    __slots__ = '_lam', '_size'
    _input_fields_ = ['_lam']

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorPoisson, self).__init__(_size=size, _state=state, _dtype=dtype,
                                            _gpu=gpu, **kw)

    def __call__(self, lam, chunks=None):
        return self.new_tensor([lam], None, raw_chunks=chunks)


def poisson(random_state, lam=1.0, size=None, chunks=None, gpu=None, **kw):
    """
    Draw samples from a Poisson distribution.

    The Poisson distribution is the limit of the binomial distribution
    for large N.

    Parameters
    ----------
    lam : float or array_like of floats
        Expectation of interval, should be >= 0. A sequence of expectation
        intervals must be broadcastable over the requested size.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``lam`` is a scalar. Otherwise,
        ``mt.array(lam).size`` samples are drawn.
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default


    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized Poisson distribution.

    Notes
    -----
    The Poisson distribution

    .. math:: f(k; \lambda)=\frac{\lambda^k e^{-\lambda}}{k!}

    For events with an expected separation :math:`\lambda` the Poisson
    distribution :math:`f(k; \lambda)` describes the probability of
    :math:`k` events occurring within the observed
    interval :math:`\lambda`.

    Because the output is limited to the range of the C long type, a
    ValueError is raised when `lam` is within 10 sigma of the maximum
    representable value.

    References
    ----------
    .. [1] Weisstein, Eric W. "Poisson Distribution."
           From MathWorld--A Wolfram Web Resource.
           http://mathworld.wolfram.com/PoissonDistribution.html
    .. [2] Wikipedia, "Poisson distribution",
           http://en.wikipedia.org/wiki/Poisson_distribution

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt
    >>> s = mt.random.poisson(5, 10000)

    Display histogram of the sample:

    >>> import matplotlib.pyplot as plt
    >>> count, bins, ignored = plt.hist(s.execute(), 14, normed=True)
    >>> plt.show()

    Draw each 100 values for lambda 100 and 500:

    >>> s = mt.random.poisson(lam=(100., 500.), size=(100, 2))
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().poisson(
            handle_array(lam), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorPoisson(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(lam, chunks=chunks)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from .. import kvstore
from ..config import options
from ..cluster_info import ClusterInfoActor
from .session import SessionManagerActor
from .resource import ResourceActor
from .kvstore import KVStoreActor
from ..node_info import NodeInfoActor


class SchedulerService(object):
    def __init__(self):
        self._cluster_info_ref = None
        self._session_manager_ref = None
        self._resource_ref = None
        self._kv_store_ref = None
        self._node_info_ref = None

    def start(self, endpoint, pool):
        """
        there are two way to start a scheduler
        1) if options.kv_store is specified as an etcd address, the endpoint will be written
        into kv-storage to indicate that this scheduler is one the schedulers,
        and the etcd is used as a service discover.
        2) if options.kv_store is not an etcd address, there will be only one scheduler
        """
        kv_store = kvstore.get(options.kv_store)
        if isinstance(kv_store, kvstore.EtcdKVStore):
            # set etcd as service discover
            service_discover_addr = options.kv_store
            schedulers = None
            kv_store.write('/schedulers/%s' % endpoint, dir=True)
        else:
            # single scheduler
            service_discover_addr = None
            schedulers = [endpoint]

        # create ClusterInfoActor
        self._cluster_info_ref = pool.create_actor(
            ClusterInfoActor, schedulers, service_discover_addr, uid=ClusterInfoActor.default_name())
        # create SessionManagerActor
        self._session_manager_ref = pool.create_actor(SessionManagerActor,
            uid=SessionManagerActor.default_name())
        # create ResourceActor
        self._resource_ref = pool.create_actor(ResourceActor, uid=ResourceActor.default_name())
        # create KVStoreActor
        self._kv_store_ref = pool.create_actor(KVStoreActor, uid=KVStoreActor.default_name())
        # create NodeInfoActor
        self._node_info_ref = pool.create_actor(NodeInfoActor, uid=NodeInfoActor.default_name())
        kv_store.write('/schedulers/%s/meta' % endpoint,
                       json.dumps(self._resource_ref.get_workers_meta()))

    def stop(self, pool):
        pool.destroy_actor(self._resource_ref)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import collections
import functools
import inspect
import json
import logging
import numbers
import os
import random
import socket
import struct
import subprocess
import sys
import time
import uuid
from collections import deque
from hashlib import md5
from datetime import date, datetime, timedelta

import numpy as np

from .compat import six, irange, functools32, getargspec
from .utils_c import *

logger = logging.getLogger(__name__)
random.seed(int(time.time()) * os.getpid())


if 'to_binary' not in globals():
    def to_binary(text, encoding='utf-8'):
        if text is None:
            return text
        if isinstance(text, six.text_type):
            return text.encode(encoding)
        elif isinstance(text, (six.binary_type, bytearray)):
            return bytes(text)
        else:
            return str(text).encode(encoding) if six.PY3 else str(text)


if 'to_text' not in globals():
    def to_text(binary, encoding='utf-8'):
        if binary is None:
            return binary
        if isinstance(binary, (six.binary_type, bytearray)):
            return binary.decode(encoding)
        elif isinstance(binary, six.text_type):
            return binary
        else:
            return str(binary) if six.PY3 else str(binary).decode(encoding)


if 'to_str' not in globals():
    def to_str(text, encoding='utf-8'):
        return to_text(text, encoding=encoding) if six.PY3 else to_binary(text, encoding=encoding)


def build_id(prefix=''):
    return prefix + '-' + str(uuid.uuid1())


# fix encoding conversion problem under windows
if sys.platform == 'win32':
    def _replace_default_encoding(func):
        def _fun(s, encoding=None):
            encoding = encoding or getattr(sys.stdout, 'encoding', None) or 'mbcs'
            return func(s, encoding=encoding)

        _fun.__name__ = func.__name__
        _fun.__doc__ = func.__doc__
        return _fun

    to_binary = _replace_default_encoding(to_binary)
    to_text = _replace_default_encoding(to_text)
    to_str = _replace_default_encoding(to_str)


if 'tokenize' not in globals():
    def tokenize(*args, **kwargs):
        try:
            import numpy as np

            def h_numpy(ob):
                if not ob.shape:
                    return str(ob), ob.dtype
                if hasattr(ob, 'mode') and getattr(ob, 'filename', None):
                    if hasattr(ob.base, 'ctypes'):
                        offset = (ob.ctypes.get_as_parameter().value -
                                  ob.base.ctypes.get_as_parameter().value)
                    else:
                        offset = 0  # root memmap's have mmap object as base
                    return (ob.filename, os.path.getmtime(ob.filename), ob.dtype,
                            ob.shape, ob.strides, offset)
                if ob.dtype.hasobject:
                    try:
                        data = md5('-'.join(ob.flat).encode('utf-8')).hexdigest()
                    except TypeError:
                        data = md5(b'-'.join([six.text_type(item).encode('utf-8') for item in
                                              ob.flat])).hexdigest()
                else:
                    try:
                        data = md5(ob.ravel().view('i1').data).hexdigest()
                    except (BufferError, AttributeError, ValueError):
                        data = md5(ob.copy().ravel().view('i1').data).hexdigest()
                return data, ob.dtype, ob.shape, ob.strides
        except ImportError:
            np = None
            h_numpy = None

        def h(ob):
            try:
                return h_non_iterative(ob)
            except TypeError:
                if isinstance(ob, dict):
                    return list, h_iterative(sorted(list(ob.items()), key=str))
                if isinstance(ob, set):
                    return list, h_iterative(sorted(ob, key=str))
                if isinstance(ob, (tuple, list)):
                    return type(ob), h_iterative(ob)
                raise TypeError('Cannot generate token for %s, type: %s' % (ob, type(ob)))

        def h_iterative(ob):
            nested = deque(ob)
            h_list = []
            dq = deque()
            while nested or dq:
                x = dq.pop() if dq else nested.popleft()
                if isinstance(x, (list, tuple)):
                    dq.extend(reversed(x))
                else:
                    h_list.append(h_non_iterative(x))
            return h_list

        def h_non_iterative(ob):
            if isinstance(ob, (int, float, str, six.text_type, bytes,
                               type(None), type, slice, date, datetime, timedelta)):
                return ob
            if isinstance(ob, six.integer_types+(complex,)):
                return ob
            if hasattr(ob, 'key'):
                return ob.key
            # numpy relative
            if h_numpy and isinstance(ob, np.ndarray):
                return h_numpy(ob)
            if np and isinstance(ob, (np.dtype, np.generic)):
                return repr(ob)

            raise TypeError

        def h_old(ob):
            if isinstance(ob, (int, float, str, six.text_type, bytes,
                               type(None), type, slice, date, datetime, timedelta)):
                return ob
            if isinstance(ob, six.integer_types+(complex,)):
                return ob
            if isinstance(ob, dict):
                return h_old(sorted(list(ob.items()), key=str))
            if isinstance(ob, (tuple, list)):
                return type(ob), list(h_old(it) for it in ob)
            if isinstance(ob, set):
                return h_old(sorted(ob, key=str))
            if hasattr(ob, 'key'):
                return ob.key
            # numpy relative
            if h_numpy and isinstance(ob, np.ndarray):
                return h_numpy(ob)
            if np and isinstance(ob, (np.dtype, np.generic)):
                return repr(ob)

            raise TypeError('Cannot generate token for %s, type: %s' % (ob, type(ob)))

        if kwargs:
            args = args + (kwargs,)
        return md5(str([h_old(arg) for arg in args]).encode('utf-8')).hexdigest()


class AttributeDict(dict):
    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(
                "'AttributeDict' object has no attribute {0}".format(item))


def hashable(obj):
    if isinstance(obj, six.string_types):
        items = obj
    elif isinstance(obj, slice):
        items = (obj.start, obj.stop, obj.step)
    elif isinstance(obj, collections.Mapping):
        items = type(obj)((k, hashable(v)) for k, v in six.iteritems(obj))
    elif isinstance(obj, collections.Iterable):
        items = tuple(hashable(item) for item in obj)
    elif isinstance(obj, collections.Hashable):
        items = obj
    elif hasattr(obj, 'key'):
        items = obj.key
    else:
        raise TypeError(type(obj))

    return items


def on_serialize_shape(shape):
    if shape:
        return tuple(s if not np.isnan(s) else -1 for s in shape)
    return shape


def on_deserialize_shape(shape):
    if shape:
        return tuple(s if s != -1 else np.nan for s in shape)
    return shape


def get_gpu_used_memory(device_id):
    import pynvml

    handle = pynvml.nvmlDeviceGetHandleByIndex(device_id)
    mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
    return mem_info.used


def parse_memory_limit(value):
    if isinstance(value, numbers.Number):
        return float(value), False
    elif value.endswith('%'):
        return float(value[:-1]) / 100, True
    elif value.lower().endswith('t'):
        return float(value[:-1]) * (1024 ** 4), False
    elif value.lower().endswith('g'):
        return float(value[:-1]) * (1024 ** 3), False
    elif value.lower().endswith('m'):
        return float(value[:-1]) * (1024 ** 2), False
    elif value.lower().endswith('k'):
        return float(value[:-1]) * 1024, False
    else:
        raise ValueError('Unknown limitation value: {0}'.format(value))


def readable_size(size):
    if size < 1024:
        return size
    elif 1024 <= size < 1024 ** 2:
        return '{0:.2f}K'.format(size / 1024)
    elif 1024 ** 2 <= size < 1024 ** 3:
        return '{0:.2f}M'.format(size / (1024 ** 2))
    elif 1024 ** 3 <= size < 1024 ** 4:
        return '{0:.2f}G'.format(size / (1024 ** 3))
    else:
        return '{0:.2f}T'.format(size / (1024 ** 4))


_commit_hash, _commit_ref = None, None


def git_info():
    from ._version import get_git_info

    global _commit_hash, _commit_ref
    if _commit_ref is not None:
        if _commit_hash is None:
            return None
        return _commit_hash, _commit_ref

    git_tuple = get_git_info()
    if git_tuple is None:
        _commit_ref, _commit_hash = ':INVALID:', None
        return None
    else:
        _commit_hash, _commit_ref = git_tuple
        return git_tuple


LOW_PORT_BOUND = 10000
HIGH_PORT_BOUND = 65535


def get_next_port(typ=None):
    import psutil
    try:
        conns = psutil.net_connections()
        typ = typ or socket.SOCK_STREAM
        occupied = set(sc.laddr.port for sc in conns
                       if sc.type == typ and LOW_PORT_BOUND <= sc.laddr.port <= HIGH_PORT_BOUND)
    except psutil.AccessDenied:
        import subprocess
        p = subprocess.Popen('netstat -a -n -p tcp'.split(), stdout=subprocess.PIPE)
        p.wait()
        occupied = set()
        for line in p.stdout:
            line = to_str(line)
            if '.' not in line:
                continue
            for part in line.split():
                if '.' in part:
                    _, port_str = part.rsplit('.', 1)
                    if port_str == '*':
                        continue
                    port = int(port_str)
                    if LOW_PORT_BOUND <= port <= HIGH_PORT_BOUND:
                        occupied.add(int(port_str))
                    break
        p.stdout.close()

    randn = struct.unpack('<Q', os.urandom(8))[0]
    idx = int(randn % (1 + HIGH_PORT_BOUND - LOW_PORT_BOUND - len(occupied)))
    for i in irange(LOW_PORT_BOUND, HIGH_PORT_BOUND + 1):
        if i in occupied:
            continue
        if idx == 0:
            return i
        idx -= 1
    raise SystemError('No ports available.')


@functools32.lru_cache(200)
def mod_hash(val, modulus):
    return int(md5(to_binary(val)).hexdigest(), 16) % modulus


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class PlasmaProcessHelper(object):
    def __init__(self, base_directory=None, proc_name=None, size=1024 ** 3,
                 socket='/tmp/plasma', one_mapped_file=False, mount_point=None, huge_pages=False):
        import pyarrow

        self._proc_name = proc_name
        self._size = size
        self._socket = socket
        self._mount_point = mount_point
        self._enable_huge_pages = huge_pages
        self._one_mapped_file = one_mapped_file

        if not base_directory:
            base_directory = pyarrow.__path__[0]
        self._base_dir = base_directory

        if proc_name is None:
            for pname in ('plasma_store', 'plasma_store_server'):
                if sys.platform == 'win32':
                    pname += '.exe'
                if os.path.exists(os.path.join(self._base_dir, pname)):
                    self._proc_name = pname
                    break

        self._process = None

    def run(self, proc_args=None):
        proc_args = proc_args or []
        if self._proc_name is None:
            raise RuntimeError('Plasma store not found.')
        args = [os.path.join(self._base_dir, self._proc_name), '-m', str(self._size),
                '-s', self._socket]
        if self._mount_point:
            args.extend(['-d', self._mount_point])
        if self._enable_huge_pages:
            args.append('-h')
        if self._one_mapped_file:
            args.append('-f')
        args.extend(proc_args)

        daemon = subprocess.Popen(args)
        logger.debug('Started %d' % daemon.pid)
        logger.debug('Params: %s' % args)
        time.sleep(2)
        self._process = daemon

    def stop(self):
        self._process.kill()


def serialize_graph(graph):
    return base64.b64encode(graph.to_pb().SerializeToString())


def deserialize_graph(graph_b64):
    from .serialize.protos.graph_pb2 import GraphDef
    from .graph import DirectedGraph
    try:
        json_obj = json.loads(to_str(graph_b64))
        return DirectedGraph.from_json(json_obj)
    except (SyntaxError, ValueError):
        g = GraphDef()
        g.ParseFromString(base64.b64decode(graph_b64))
        return DirectedGraph.from_pb(g)


def merge_tensor_chunks(input_tensor, ctx):
    from .tensor.execution.core import Executor
    from .tensor.expressions.datasource import TensorFetchChunk

    if len(input_tensor.chunks) == 1:
        return ctx[input_tensor.chunks[0].key]

    chunks = []
    for c in input_tensor.chunks:
        op = TensorFetchChunk(dtype=c.dtype, to_fetch_key=c.key)
        chunk = op.new_chunk(None, c.shape, index=c.index, _key=c.key)
        chunks.append(chunk)

    new_op = TensorFetchChunk(dtype=input_tensor.dtype, to_fetch_key=input_tensor.key)
    tensor = new_op.new_tensor(None, input_tensor.shape, chunks=chunks,
                               nsplits=input_tensor.nsplits)

    executor = Executor(storage=ctx)
    concat_result = executor.execute_tensor(tensor, concat=True)
    return concat_result[0]


if sys.version_info[0] < 3:
    def wraps(fun):
        if isinstance(fun, functools.partial):
            return lambda f: f
        return functools.wraps(fun)
else:
    wraps = functools.wraps


def calc_data_size(dt):
    if isinstance(dt, tuple):
        return sum(c.nbytes for c in dt)
    else:
        return dt.nbytes


def log_unhandled(func):
    frame_globals = inspect.currentframe().f_back.f_globals
    mod_logger = None
    for logger_name in ('logger', 'LOG', 'LOGGER'):
        if logger_name in frame_globals:
            mod_logger = frame_globals[logger_name]
            break
    if not mod_logger:
        return func

    func_name = getattr(func, '__qualname__', func.__module__ + func.__name__)
    func_args = getargspec(func)

    @wraps(func)
    def _wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            kwcopy = kwargs.copy()
            kwcopy.update(zip(func_args.args, args))

            messages = []
            for k, v in kwcopy.items():
                if 'key' in k:
                    messages.append('%s=%r' % (k, v))

            err_msg = 'Unexpected exception occurred in %s.' % func_name
            if messages:
                err_msg += ' ' + ' '.join(messages)
            mod_logger.exception(err_msg)
            raise
    return _wrapped



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..utils import validate_axis


def rollaxis(tensor, axis, start=0):
    """
    Roll the specified axis backwards, until it lies in a given position.

    This function continues to be supported for backward compatibility, but you
    should prefer `moveaxis`.

    Parameters
    ----------
    a : Tensor
        Input tensor.
    axis : int
        The axis to roll backwards.  The positions of the other axes do not
        change relative to one another.
    start : int, optional
        The axis is rolled until it lies before this position.  The default,
        0, results in a "complete" roll.

    Returns
    -------
    res : Tensor
        a view of `a` is always returned.

    See Also
    --------
    moveaxis : Move array axes to new positions.
    roll : Roll the elements of an array by a number of positions along a
        given axis.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.ones((3,4,5,6))
    >>> mt.rollaxis(a, 3, 1).shape
    (3, 6, 4, 5)
    >>> mt.rollaxis(a, 2).shape
    (5, 3, 4, 6)
    >>> mt.rollaxis(a, 1, 4).shape
    (3, 5, 6, 4)

    """
    n = tensor.ndim
    axis = validate_axis(n, axis)
    if start < 0:
        start += n
    msg = "'%s' arg requires %d <= %s < %d, but %d was passed in"
    if not (0 <= start < n + 1):
        raise np.AxisError(msg % ('start', -n, 'start', n + 1, start))
    if axis < start:
        # it's been removed
        start -= 1
    if axis == start:
        return tensor
    axes = list(range(0, n))
    axes.remove(axis)
    axes.insert(start, axis)
    return tensor.transpose(axes)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .server import register_ui_handler, get_jinja_env, MarsWebAPI


def dashboard(scheduler_ip, doc):
    doc.title = 'Mars UI'

    web_api = MarsWebAPI(scheduler_ip)
    infos = web_api.get_schedulers_info()
    doc.template_variables['infos'] = infos
    jinja_env = get_jinja_env()
    doc.template = jinja_env.get_template('dashboard.html')


register_ui_handler('/', dashboard)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import os
import sys

from .actors import create_actor_pool
from .config import options
from .errors import StartArgumentError
from .utils import get_next_port

logger = logging.getLogger(__name__)


class BaseApplication(object):
    """
    :type pool mars.actors.pool.gevent_pool.ActorContext
    """
    service_description = ''
    service_logger = logger

    def __init__(self):
        self.args = None
        self.endpoint = None
        self.pool = None
        self.n_process = None

        self._running = False

    def __call__(self, argv=None):
        import json

        if argv is None:
            argv = sys.argv[1:]
        new_argv = []
        for a in argv:
            if not a.startswith('-D'):
                new_argv.append(a)
                continue
            conf, val = a[2:].split('=', 1)
            conf_parts = conf.split('.')
            conf_obj = options
            for g in conf_parts[:-1]:
                conf_obj = getattr(conf_obj, g)
            try:
                setattr(conf_obj, conf_parts[-1], json.loads(val))
            except:
                setattr(conf_obj, conf_parts[-1], val)

        return self._main(new_argv)

    def _main(self, argv=None):
        parser = argparse.ArgumentParser(description=self.service_description)
        parser.add_argument('-a', '--advertise', help='advertise ip')
        parser.add_argument('-k', '--kv-store', help='address of kv store service, for instance, etcd')
        parser.add_argument('-e', '--endpoint', help='endpoint of the service')
        parser.add_argument('-H', '--host', help='host of the scheduler service, only available '
                                                 'when `endpoint` is absent')
        parser.add_argument('-p', '--port', help='port of the scheduler service, only available '
                                                 'when `endpoint` is absent')
        parser.add_argument('--level', help='log level')
        parser.add_argument('--format', help='log format')
        parser.add_argument('--log_conf', help='log config file')
        parser.add_argument('--inspect', help='inspection endpoint')
        parser.add_argument('--load-modules', nargs='*', help='modules to import')
        parser.add_argument('--profile', nargs='?', default=-1, help='profile application')
        self.config_args(parser)
        args = parser.parse_args(argv)
        self.args = args

        endpoint = args.endpoint
        host = args.host
        port = args.port
        if args.kv_store:
            options.kv_store = args.kv_store

        load_modules = []
        for mod in args.load_modules or ():
            load_modules.extend(mod.split(','))
        if not args.load_modules:
            load_module_str = os.environ.get('MARS_LOAD_MODULES')
            if load_module_str:
                load_modules = load_module_str.split(',')
        load_modules.append('mars.tensor')
        [__import__(m, globals(), locals(), []) for m in load_modules]
        if load_modules:
            self.service_logger.info('Modules %s loaded', ','.join(load_modules))

        self.n_process = 1

        self.config_service()
        self.config_logging()

        if not host:
            host = args.advertise or '0.0.0.0'
        if not endpoint and port:
            endpoint = host + ':' + port

        try:
            self.validate_arguments()
        except StartArgumentError as ex:
            parser.error('Failed to start application: %s' % ex)

        if getattr(self, 'require_pool', True):
            self.endpoint, self.pool = self._try_create_pool(endpoint=endpoint, host=host, port=port)
        self.service_logger.info('%s started at %s.', self.service_description, self.endpoint)
        self.main_loop()

    def config_logging(self):
        import logging.config
        log_conf = self.args.log_conf or 'logging.conf'

        conf_file_paths = [
            '', os.path.abspath('.'),
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        ]
        log_configured = False
        for path in conf_file_paths:
            conf_path = log_conf
            if path:
                conf_path = os.path.join(conf_path)
            if os.path.exists(conf_path):
                logging.config.fileConfig(conf_path, disable_existing_loggers=False)
                log_configured = True

        if not log_configured:
            if not self.args.level:
                level = logging.INFO
            else:
                level = getattr(logging, self.args.level.upper())
            logging.basicConfig(level=level, format=self.args.format)

    def validate_arguments(self):
        pass

    def _try_create_pool(self, endpoint=None, host=None, port=None):
        pool = None
        if endpoint:
            pool = self.create_pool(address=endpoint)
        else:
            use_port = None
            retrial = 5
            while use_port is None:
                use_port = port or get_next_port()
                try:
                    endpoint = '{0}:{1}'.format(host, use_port)
                    pool = self.create_pool(address=endpoint)
                    break
                except:
                    retrial -= 1
                    if retrial == 0:
                        raise

                    if port is None:
                        use_port = None
                    else:
                        raise
        return endpoint, pool

    def create_pool(self, *args, **kwargs):
        kwargs.update(dict(n_process=self.n_process, backend='gevent'))
        return create_actor_pool(*args, **kwargs)

    def main_loop(self):
        if self.args.profile == -1:
            profile_file = None
        else:
            profile_file = self.args.profile or ('mars_' + self.__class__.__name__ + '.prof')
        try:
            if profile_file:
                import yappi
                yappi.set_clock_type('wall')
                yappi.start(builtins=False, profile_threads=False)

            with self.pool:
                try:
                    self.start_service()
                    self._running = True
                    while True:
                        self.pool.join(1)
                        for proc in self.pool.processes:
                            if not proc.is_alive():
                                self.service_logger.fatal(
                                    'Process %d exited unpredictably. exitcode=%d', proc.pid, proc.exitcode)
                                raise KeyboardInterrupt
                except:
                    self._running = False
                    self.stop_service()
        finally:
            self._running = False
            if profile_file:
                import yappi
                yappi.logger = logging.getLogger(__name__)
                p = yappi.convert2pstats(yappi.get_func_stats())
                p.strip_dirs()
                p.sort_stats('time')
                p.print_stats(40)
                p.dump_stats(profile_file)

    def config_service(self):
        pass

    def config_args(self, parser):
        raise NotImplementedError

    def start_service(self):
        raise NotImplementedError

    def stop_service(self):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..datasource import tensor as astensor
from .dot import dot


def vdot(a, b):
    """
    Return the dot product of two vectors.

    The vdot(`a`, `b`) function handles complex numbers differently than
    dot(`a`, `b`).  If the first argument is complex the complex conjugate
    of the first argument is used for the calculation of the dot product.

    Note that `vdot` handles multidimensional tensors differently than `dot`:
    it does *not* perform a matrix product, but flattens input arguments
    to 1-D vectors first. Consequently, it should only be used for vectors.

    Parameters
    ----------
    a : array_like
        If `a` is complex the complex conjugate is taken before calculation
        of the dot product.
    b : array_like
        Second argument to the dot product.

    Returns
    -------
    output : Tensor
        Dot product of `a` and `b`.  Can be an int, float, or
        complex depending on the types of `a` and `b`.

    See Also
    --------
    dot : Return the dot product without using the complex conjugate of the
          first argument.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([1+2j,3+4j])
    >>> b = mt.array([5+6j,7+8j])
    >>> mt.vdot(a, b).execute()
    (70-8j)
    >>> mt.vdot(b, a).execute()
    (70+8j)

    Note that higher-dimensional arrays are flattened!

    >>> a = mt.array([[1, 4], [5, 6]])
    >>> b = mt.array([[4, 1], [2, 2]])
    >>> mt.vdot(a, b).execute()
    30
    >>> mt.vdot(b, a).execute()
    30
    >>> 1*4 + 4*1 + 5*2 + 6*2
    30
    """
    a, b = astensor(a), astensor(b)
    return dot(a.conj().ravel(), b.ravel())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import AnyField
from .core import TensorNoInput


class Scalar(TensorNoInput):
    """
    Operand represents scalar type.
    """

    _op_type_ = OperandDef.SCALAR

    _data = AnyField('data')

    def __init__(self, data=None, dtype=None, gpu=False, **kw):
        super(Scalar, self).__init__(_data=data, _dtype=dtype, _gpu=gpu, **kw)

    @classmethod
    def tile(cls, op):
        chunk_op = op.copy().reset_key()
        chunk = chunk_op.new_chunk(None, (), index=())
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape,
                                  chunks=[chunk], nsplits=())

    @property
    def data(self):
        return self._data


def scalar(data, dtype=None, gpu=False):
    try:
        arr = np.array(data, dtype=dtype)
        op = Scalar(np.asscalar(arr), dtype=arr.dtype, gpu=gpu)
        shape = ()
        return op(shape)
    except ValueError:
        raise TypeError('Expect scalar, got: {0}'.format(data))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorUnaryOp


class TensorFloor(operands.Floor, TensorUnaryOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorFloor, self).__init__(_casting=casting, _err=err,
                                          _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x):
        return x.issparse()


@infer_dtype(np.floor)
def floor(x, out=None, where=None, **kwargs):
    """
    Return the floor of the input, element-wise.

    The floor of the scalar `x` is the largest integer `i`, such that
    `i <= x`.  It is often denoted as :math:`\lfloor x \rfloor`.

    Parameters
    ----------
    x : array_like
        Input data.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor or scalar
        The floor of each element in `x`.

    See Also
    --------
    ceil, trunc, rint

    Notes
    -----
    Some spreadsheet programs calculate the "floor-towards-zero", in other
    words ``floor(-2.5) == -2``.  NumPy instead uses the definition of
    `floor` where `floor(-2.5) == -3`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([-1.7, -1.5, -0.2, 0.2, 1.5, 1.7, 2.0])
    >>> mt.floor(a).execute()
    array([-2., -2., -1.,  0.,  1.,  1.,  2.])
    """
    op = TensorFloor(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorArgReduction


class TensorNanArgmaxChunk(operands.NanArgmaxChunk, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None,
                 combine_size=None, offset=None, total_shape=None,**kw):
        super(TensorNanArgmaxChunk, self).__init__(_axis=axis, _dtype=dtype,
                                                   _keepdims=keepdims, _combine_size=combine_size,
                                                   _offset=offset, _total_shape=total_shape, **kw)


class TensorNanArgmaxCombine(operands.NanArgmaxCombine, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None, combine_size=None, **kw):
        super(TensorNanArgmaxCombine, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                                     _combine_size=combine_size, **kw)


class TensorNanArgmax(operands.NanArgmax, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None, combine_size=None, **kw):
        super(TensorNanArgmax, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                              _combine_size=combine_size, **kw)

    @staticmethod
    def _get_op_types():
        return TensorNanArgmaxChunk, TensorNanArgmax, TensorNanArgmaxCombine


def nanargmax(a, axis=None, out=None, keepdims=None, combine_size=None):
    """
    Return the indices of the maximum values in the specified axis ignoring
    NaNs. For all-NaN slices ``ValueError`` is raised. Warning: the
    results cannot be trusted if a slice contains only NaNs and -Infs.


    Parameters
    ----------
    a : array_like
        Input data.
    axis : int, optional
        Axis along which to operate.  By default flattened input is used.
    out : Tensor, optional
        Alternate output tensor in which to place the result.  The default
        is ``None``; if provided, it must have the same shape as the
        expected output, but the type will be cast if necessary.
        See `doc.ufuncs` for details.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the input tensor.

        If the default value is passed, then `keepdims` will not be
        passed through to the `mean` method of sub-classes of
        `Tensor`, however any non-default value will be.  If the
        sub-classes `sum` method does not implement `keepdims` any
        exceptions will be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    index_array : Tensor
        An tensor of indices or a single index value.

    See Also
    --------
    argmax, nanargmin

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([[mt.nan, 4], [2, 3]])
    >>> mt.argmax(a).execute()
    0
    >>> mt.nanargmax(a).execute()
    1
    >>> mt.nanargmax(a, axis=0).execute()
    array([1, 0])
    >>> mt.nanargmax(a, axis=1).execute()
    array([1, 1])

    """
    op = TensorNanArgmax(axis=axis, dtype=np.dtype(int), keepdims=keepdims, combine_size=combine_size)
    return op(a, out=out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Iterable

from ..utils import validate_axis, decide_chunks, recursive_tile
from ..core import TensorOperandMixin


class TensorFFTMixin(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def _get_shape(cls, op, shape):
        raise NotImplementedError

    @classmethod
    def tile(cls, op):
        in_tensor = op.inputs[0]
        axis = op.axis

        if in_tensor.chunk_shape[axis] != 1:
            # fft requires only 1 chunk for the specified axis, so we do rechunk first
            chunks = {validate_axis(in_tensor.ndim, axis): in_tensor.shape[axis]}
            new_chunks = decide_chunks(in_tensor.shape, chunks, in_tensor.dtype.itemsize)
            in_tensor = in_tensor.rechunk(new_chunks).single_tiles()

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk_shape = cls._get_shape(op, c.shape)
            out_chunk = chunk_op.new_chunk([c], chunk_shape, index=c.index)
            out_chunks.append(out_chunk)

        nsplits = [tuple(c.shape[i] for c in out_chunks
                         if all(idx == 0 for j, idx in enumerate(c.index) if j != i))
                   for i in range(len(out_chunks[0].shape))]
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape,
                                  chunks=out_chunks, nsplits=nsplits)


def validate_fft(tensor, axis=-1, norm=None):
    validate_axis(tensor.ndim, axis)
    if norm is not None and norm not in ('ortho',):
        raise ValueError('Invalid norm value {0}, should be None or "ortho"'.format(norm))


class TensorFFTNMixin(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def _get_shape(cls, op, shape):
        raise NotImplementedError

    @classmethod
    def tile(cls, op):
        in_tensor = op.inputs[0]
        axes = op.axes

        if any(in_tensor.chunk_shape[axis] != 1 for axis in axes):
            new_chunks = decide_chunks(
                in_tensor.shape, {validate_axis(in_tensor.ndim, axis): in_tensor.shape[axis] for axis in axes},
                in_tensor.dtype.itemsize)
            in_tensor = in_tensor.rechunk(new_chunks).single_tiles()

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk_shape = cls._get_shape(op, c.shape)
            out_chunk = chunk_op.new_chunk([c], chunk_shape, index=c.index)
            out_chunks.append(out_chunk)

        nsplits = [tuple(c.shape[i] for c in out_chunks
                         if all(idx == 0 for j, idx in enumerate(c.index) if j != i))
                   for i in range(len(out_chunks[0].shape))]
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape,
                                  chunks=out_chunks, nsplits=nsplits)


def validate_fftn(tensor, s=None, axes=None, norm=None):
    if axes is None:
        if s is None:
            axes = tuple(range(tensor.ndim))
        else:
            axes = tuple(range(len(s)))
    else:
        [validate_axis(tensor.ndim, axis) for axis in axes]
        if len(set(axes)) < len(axes):
            raise ValueError('Duplicate axes not allowed')

    if norm is not None and norm not in ('ortho',):
        raise ValueError('Invalid norm value {0}, should be None or "ortho"'.format(norm))

    return axes


class TensorFFTShiftMixin(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def _is_inverse(cls):
        return False

    @classmethod
    def _process_axes(cls, x, axes):
        if axes is None:
            axes = tuple(range(x.ndim))
        elif isinstance(axes, Iterable):
            axes = tuple(axes)
        else:
            axes = (axes,)

        return axes

    @classmethod
    def tile(cls, op):
        from ..merge import concatenate

        axes = op.axes
        in_tensor = op.input
        is_inverse = cls._is_inverse()

        x = in_tensor
        for axis in axes:
            size = in_tensor.shape[axis]
            slice_on = (size + 1) // 2 if not is_inverse else size // 2
            slc1 = [slice(None)] * axis + [slice(slice_on)]
            slc2 = [slice(None)] * axis + [slice(slice_on, None)]
            x = concatenate([x[slc2], x[slc1]], axis=axis)

        recursive_tile(x)
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape,
                                  chunks=x.chunks, nsplits=x.nsplits)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorLogseries(operands.Logseries, TensorRandomOperandMixin):
    __slots__ = '_p', '_size'
    _input_fields_ = ['_p']

    def __init__(self, state=None, size=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorLogseries, self).__init__(_state=state, _size=size, _dtype=dtype,
                                              _gpu=gpu, **kw)

    def __call__(self, p, chunks=None):
        return self.new_tensor([p], None, raw_chunks=chunks)


def logseries(random_state, p, size=None, chunks=None, gpu=None, **kw):
    """
    Draw samples from a logarithmic series distribution.

    Samples are drawn from a log series distribution with specified
    shape parameter, 0 < ``p`` < 1.

    Parameters
    ----------
    p : float or array_like of floats
        Shape parameter for the distribution.  Must be in the range (0, 1).
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``p`` is a scalar.  Otherwise,
        ``np.array(p).size`` samples are drawn.
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized logarithmic series distribution.

    See Also
    --------
    scipy.stats.logser : probability density function, distribution or
        cumulative density function, etc.

    Notes
    -----
    The probability density for the Log Series distribution is

    .. math:: P(k) = \frac{-p^k}{k \ln(1-p)},

    where p = probability.

    The log series distribution is frequently used to represent species
    richness and occurrence, first proposed by Fisher, Corbet, and
    Williams in 1943 [2].  It may also be used to model the numbers of
    occupants seen in cars [3].

    References
    ----------
    .. [1] Buzas, Martin A.; Culver, Stephen J.,  Understanding regional
           species diversity through the log series distribution of
           occurrences: BIODIVERSITY RESEARCH Diversity & Distributions,
           Volume 5, Number 5, September 1999 , pp. 187-195(9).
    .. [2] Fisher, R.A,, A.S. Corbet, and C.B. Williams. 1943. The
           relation between the number of species and the number of
           individuals in a random sample of an animal population.
           Journal of Animal Ecology, 12:42-58.
    .. [3] D. J. Hand, F. Daly, D. Lunn, E. Ostrowski, A Handbook of Small
           Data Sets, CRC Press, 1994.
    .. [4] Wikipedia, "Logarithmic distribution",
           http://en.wikipedia.org/wiki/Logarithmic_distribution

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt
    >>> import matplotlib.pyplot as plt

    >>> a = .6
    >>> s = mt.random.logseries(a, 10000)
    >>> count, bins, ignored = plt.hist(s.execute())

    #   plot against distribution

    >>> def logseries(k, p):
    ...     return -p**k/(k*mt.log(1-p))
    >>> plt.plot(bins, (logseries(bins, a)*count.max()/
    ...          logseries(bins, a).max()).execute(), 'r')
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().logseries(
            handle_array(p), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorLogseries(state=random_state._state, size=size, gpu=gpu, **kw)
    return op(p, chunks=chunks)
